#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""The AWIPS Tiled writer is used to create AWIPS-compatible tiled NetCDF4 files.

The Advanced Weather Interactive Processing System (AWIPS) is a
program used by the United States National Weather Service (NWS) and others
to view
different forms of weather imagery. The original Sectorized Cloud and Moisture
Imagery (SCMI) functionality in AWIPS was a NetCDF4 format supported by AWIPS
to store one image broken up in to one or more "tiles". This format has since
been expanded to support many other products and so the writer for this format
in Satpy is generically called the "AWIPS Tiled" writer. You may still see
SCMI referenced in this documentation or in the source code for the writer.
Once AWIPS is configured for specific products this writer can be used to
provide compatible products to the system.

The AWIPS Tiled writer takes 2D (y, x) geolocated data and creates one or more
AWIPS-compatible NetCDF4 files. The writer and the AWIPS client may
need to be configured to make things appear the way the user wants in
the AWIPS client. The writer can only produce files for datasets mapped
to areas with specific projections:

 - lcc
 - geos
 - merc
 - stere

This is a limitation of the AWIPS client and not of the writer. In the case
where AWIPS has been updated to support additional projections, this writer
may also need to be updated to support those projections.

AWIPS Configuration
-------------------

Depending on how this writer is used and the data it is provided, AWIPS may
need additional configuration on the server side to properly ingest the files
produced. This will require administrator privileges to the ingest server(s)
and is not something that can be configured on the client. Note that any
changes required must be done on all servers that you wish to ingest your data
files. The generic "polar" template this writer defaults to should limit the
number of modifications needed for any new data fields that AWIPS previously
was unaware of. Once the data is ingested, the client can be used to customize
how the data looks on screen.

AWIPS requires files to follow a specific naming scheme so they can be routed
to specific "decoders". For the files produced by this writer, this typically
means editing the "goesr" decoder configuration in a directory like::

/awips2/edex/data/utility/common_static/site/<site>/distribution/goesr.xml

The "goesr" decoder is a subclass of the "satellite" decoder. You may see
either name show up in the AWIPS ingest logs. With the correct
regular expression in the above file, your files should be passed to the
right decoder, opened, and parsed for data.

To tell AWIPS exactly what attributes and variables mean in your file, you'll
need to create or configure an XML file in::

/awips2/edex/data/utility/common_static/site/<site>/satellite/goesr/descriptions/

See the existing files in this directory for examples. The "polar" template
(see below) that this writer uses by default is already configured in the
"Polar" subdirectory assuming that the TOWR-S RPM package has been installed
on your AWIPS ingest server.

Templates
---------

This writer allows for a "template" to be specified to control how the output
files are structured and created. Templates can be configured in the writer
YAML file (``awips_tiled.yaml``) or passed as a dictionary to the ``template``
keyword argument. Templates have three main sections:

1. global_attributes
2. coordinates
3. variables

Additionally, you can specify whether a template should produce files with
one variable per file by specifying ``single_variable: true`` or multiple
variables per file by specifying ``single_variable: false``. You can also
specify the output filename for a template using a Python format string.
See ``awips_tiled.yaml`` for examples. Lastly, a ``add_sector_id_global``
boolean parameter can be specified to add the user-provided ``sector_id``
keyword argument as a global attribute to the file.

The ``global_attributes`` section takes names of global attributes and
then a series of options to "render" that attribute from the metadata
provided when creating files. For example::

    product_name:
        value: "{name}"

For more information see the
:meth:`satpy.writers.awips_tiled.NetCDFTemplate.get_attr_value` method.

The ``coordinates`` and ``variables`` are similar to each other in that they
define how a variable should be created, the attributes it should have, and
the encoding to write to the file. Coordinates typically don't need to be
modified as tiled files usually have only ``x`` and ``y`` dimension variables.
The Variables on the other hand use a decision tree to determine what section
applies for a particular DataArray being saved. The basic structure is::

    variables:
      arbitrary_section_name:
        <decision tree matching parameters>
        var_name: "output_netcdf_variable_name"
        attributes:
          <attributes similar to global attributes>
        encoding:
          <xarray encoding parameters>

The "decision tree matching parameters" can be one or more of "name",
"standard_name', "satellite", "sensor", "area_id', "units", or "reader".
The writer will choose the best section for the DataArray being saved
(the most matches). If none of these parameters are specified in a section
then it will be used when no other matches are found (the "default" section).

The "encoding" parameters can be anything accepted by xarray's ``to_netcdf``
method. See :meth:`xarray.Dataset.to_netcdf` for more information on the
`encoding`` keyword argument.

For more examples see the existing builtin templates defined in
``awips_tiled.yaml``.

Builtin Templates
^^^^^^^^^^^^^^^^^

There are only a few templates provided in Sapty currently.

* **polar**: A custom format developed for the CSPP Polar2Grid project at the
  University of Wisconsin - Madison Space Science and Engineering Center
  (SSEC). This format is made available through the TOWR-S package that can be
  installed for GOES-R support in AWIPS. This format is meant to be very
  generic and should theoretically allow any variable to get ingested into
  AWIPS.
* **glm_l2_radc**: This format is used to produce standard files for the gridded
  GLM products produced by the CSPP Geo Gridded GLM package. Support for this
  format is also available in the TOWR-S package on an AWIPS ingest server.
  This format is specific to gridded GLM on the CONUS sector and is not meant
  to work for other data.
* **glm_l2_radf**: This format is used to produce standard files for the gridded
  GLM productes produced by the CSPP Geo Gridded GLM package. Support for this
  format is also available in the TOWR-S package on an AWIPS ingest server.
  This format is specific to gridded GLM on the Full Disk sector and is not
  meant to work for other data.

Numbered versus Lettered Grids
------------------------------

By default this writer will save tiles by number starting with '1'
representing the upper-left image tile. Tile numbers then increase
along the column and then on to the next row.

By specifying `lettered_grid` as `True` tiles can be designated with a
letter. Lettered grids or sectors are preconfigured in the `awips_tiled.yaml`
configuration file. The lettered tile locations are static and will not
change with the data being written to them. Each lettered tile is split
into a certain number of subtiles (`num_subtiles`), default 2 rows by
2 columns. Lettered tiles are meant to make it easier for receiving
AWIPS clients/stations to filter what tiles they receive; saving time,
bandwidth, and space.

Any tiles (numbered or lettered) not containing any valid data are not
created.

Updating tiles
--------------

There are some input data cases where we want to put new data in a tile
file written by a previous execution. An example is a pre-tiled input dataset
that is processed one tile at a time. One input tile may map to one or more
output AWIPS tiles, but may not perfectly aligned, leaving
empty/unused space in the output tile. The next input tile may be able to fill
in that empty space and should be allowed to write the "new" data to the file.
This is the default behavior of the AWIPS tiled writer. In cases where data
overlaps the existing data in the tile, the newer data has priority.

Shifting Lettered Grids
-----------------------

Due to the static nature of the lettered grids, there is sometimes a
need to shift the locations of where these tiles are by up to 0.5 pixels in
each dimension to align with the data being processed. This means that the
tiles for a 1000m resolution grid may be shifted up to 500m in each direction
from the original definition of the lettered "sector". This can cause
differences in the location of the tiles between executions depending on the
locations of the input data. In the worst case tile A01 from one execution
could be shifted up to 1 grid cell from tile A01 in another execution (one
is shifted 0.5 pixels to the left, the other is shifted 0.5 to the right).

This shifting makes the calculations for generating tiles easier and
more accurate. By default, the lettered tile locations are changed to match
the location of the data. This works well when output tiles will not be
updated (see above) in future processing. In cases where output tiles will be
filled in or updated with more data the ``use_sector_reference`` keyword
argument can be set to ``True`` to tell the writer to shift the data's
geolocation by up to 0.5 pixels in each dimension instead of shifting the
lettered tile locations.

"""

import datetime as dt
import logging
import os
import string
import sys
import warnings
from collections import namedtuple
from typing import Any

import dask
import dask.array as da
import numpy as np
import xarray as xr
from pyproj import CRS, Proj, Transformer
from pyresample.geometry import AreaDefinition
from trollsift.parser import Parser, StringFormatter

from satpy import __version__
from satpy.decision_tree import DecisionTree
from satpy.writers.core.base import Writer

LOG = logging.getLogger(__name__)
DEFAULT_OUTPUT_PATTERN = "{source_name}_AII_{platform_name}_{sensor}_" \
                         "{name}_{sector_id}_{tile_id}_" \
                         "{start_time:%Y%m%d_%H%M}.nc"

UNIT_CONV = {
    "micron": "microm",
    "mm h-1": "mm/h",
    "1": "*1",
    "none": "*1",
    "percent": "%",
    "Kelvin": "kelvin",
    "K": "kelvin",
    "Meter": "meters",
}

TileInfo = namedtuple("TileInfo", ["tile_count", "image_shape", "tile_shape",
                                   "tile_row_offset", "tile_column_offset", "tile_id",
                                   "tile_number",
                                   "x", "y", "xy_factors", "tile_slices", "data_slices"])
XYFactors = namedtuple("XYFactors", ["mx", "bx", "my", "by"])


class NumberedTileGenerator(object):
    """Helper class to generate per-tile metadata for numbered tiles."""

    def __init__(self, area_definition,
                 tile_shape=None, tile_count=None):
        """Initialize and generate tile information for this sector/grid for later use."""
        self.area_definition = area_definition
        self._rows = self.area_definition.height
        self._cols = self.area_definition.width

        # get tile shape, number of tiles, etc.
        self._get_tile_properties(tile_shape, tile_count)
        # scaling parameters for the overall images X and Y coordinates
        # they must be the same for all X and Y variables for all tiles
        # and must be stored in the file as 0, 1, 2, 3, ...
        # (X factor, X offset, Y factor, Y offset)
        self.mx, self.bx, self.my, self.by = self._get_xy_scaling_parameters()
        self.xy_factors = XYFactors(self.mx, self.bx, self.my, self.by)
        self._tile_cache = []

    def _get_tile_properties(self, tile_shape, tile_count):
        """Generate tile information for numbered tiles."""
        if tile_shape is not None:
            tile_shape = (int(min(tile_shape[0], self._rows)), int(min(tile_shape[1], self._cols)))
            tile_count = (int(np.ceil(self._rows / float(tile_shape[0]))),
                          int(np.ceil(self._cols / float(tile_shape[1]))))
        elif tile_count:
            tile_shape = (int(np.ceil(self._rows / float(tile_count[0]))),
                          int(np.ceil(self._cols / float(tile_count[1]))))
        else:
            raise ValueError("Either 'tile_count' or 'tile_shape' must be provided")

        # number of pixels per each tile (rows, cols)
        self.tile_shape = tile_shape
        # number of tiles in each direction (rows, columns)
        self.tile_count = tile_count
        # number of tiles in the entire image
        self.total_tiles = tile_count[0] * tile_count[1]
        # number of pixels in the whole image (rows, columns)
        self.image_shape = (self.tile_shape[0] * self.tile_count[0],
                            self.tile_shape[1] * self.tile_count[1])

        # X and Y coordinates of the whole image
        self.x, self.y = self._get_xy_arrays()

    def _get_xy_arrays(self):
        """Get the overall X/Y coordinate variable arrays."""
        gd = self.area_definition
        ts = self.tile_shape
        tc = self.tile_count
        # Since our tiles may go over the edge of the original "grid" we
        # need to make sure we calculate X/Y to the edge of all of the tiles
        imaginary_data_size = (ts[0] * tc[0], ts[1] * tc[1])
        ps_x = gd.pixel_size_x
        ps_y = gd.pixel_size_y
        # tiles start from upper-left
        new_extents = (
            gd.area_extent[0],
            gd.area_extent[1] - ps_y * (imaginary_data_size[0] - gd.height),
            gd.area_extent[2] + ps_x * (imaginary_data_size[1] - gd.width),
            gd.area_extent[3])
        imaginary_grid_def = AreaDefinition(
            gd.area_id,
            gd.description,
            gd.proj_id,
            gd.crs,
            imaginary_data_size[1],
            imaginary_data_size[0],
            new_extents,
        )

        x, y = imaginary_grid_def.get_proj_vectors()
        return x, y

    def _get_xy_scaling_parameters(self):
        """Get the X/Y coordinate limits for the full resulting image."""
        gd = self.area_definition
        bx = self.x.min()
        mx = gd.pixel_size_x
        by = self.y.max()
        my = -abs(gd.pixel_size_y)
        return mx, bx, my, by

    def _tile_number(self, ty, tx):
        """Get tile number from tile row/column."""
        # e.g.
        # 001 002 003 004
        # 005 006 ...
        return ty * self.tile_count[1] + tx + 1

    def _tile_identifier(self, ty, tx):
        """Get tile identifier for numbered tiles."""
        return "T{:03d}".format(self._tile_number(ty, tx))

    def _generate_tile_info(self):
        """Get numbered tile metadata."""
        x = self.x
        y = self.y
        ts = self.tile_shape
        tc = self.tile_count

        if self._tile_cache:
            for tile_info in self._tile_cache:
                yield tile_info

        for ty in range(tc[0]):
            for tx in range(tc[1]):
                tile_id = self._tile_identifier(ty, tx)
                tile_row_offset = ty * ts[0]
                tile_column_offset = tx * ts[1]

                # store tile data to an intermediate array
                # the tile may be larger than the remaining data, handle that:
                max_row_idx = min((ty + 1) * ts[0], self._rows) - (ty * ts[0])
                max_col_idx = min((tx + 1) * ts[1], self._cols) - (tx * ts[1])
                tile_slices = (slice(0, max_row_idx), slice(0, max_col_idx))
                data_slices = (slice(ty * ts[0], (ty + 1) * ts[0]),
                               slice(tx * ts[1], (tx + 1) * ts[1]))

                tmp_x = x[data_slices[1]]
                tmp_y = y[data_slices[0]]

                tile_number = self._tile_number(ty, tx)
                tile_info = TileInfo(
                    tc, self.image_shape, ts,
                    tile_row_offset, tile_column_offset, tile_id,
                    tile_number,
                    tmp_x, tmp_y, self.xy_factors, tile_slices, data_slices)
                self._tile_cache.append(tile_info)
                yield tile_info

    def __call__(self):
        """Provide simple call interface for getting tile metadata."""
        if self._tile_cache:
            tile_infos = self._tile_cache
        else:
            tile_infos = self._generate_tile_info()

        for tile_info in tile_infos:
            # TODO: Return the slice instead of the actual data array
            #   Use the slicing start/end to determine if it is empty
            # tile_data = data[tile_info.data_slices]
            # if not tile_data.size:
            #     LOG.info("Tile {} is empty, skipping...".format(tile_info[2]))
            #     continue
            yield tile_info


class LetteredTileGenerator(NumberedTileGenerator):
    """Helper class to generate per-tile metadata for lettered tiles."""

    def __init__(
            self,
            area_definition: AreaDefinition,
            extents: tuple[float, float, float, float],
            sector_crs: CRS,
            cell_size: tuple[int, int] = (2000000, 2000000),
            num_subtiles: tuple[int, int] | None = None,
            use_sector_reference: bool = False):
        """Initialize tile information for later generation.

        Args:
            area_definition: Area of the data being saved.
            extents: Four element tuple of the configured lettered area.
            sector_crs: CRS of the configured lettered sector area.
            cell_size: Two element tuple of resolution of each tile
                in sector projection units (y, x).
            num_subtiles: Two element tuple of number of sub-tiles to create
                for each larger lettered tile. Defaults to (2, 2).
            use_sector_reference: If False (default), the data's geolocation
                is used to determine where pixels should align. If True, the
                data's geolocation is shifted by up to half a pixel to align
                with the extents of the lettered grid/area.

        """
        # (row subtiles, col subtiles)
        self.num_subtiles = num_subtiles or (2, 2)
        self.cell_size = cell_size  # (row tile height, col tile width)
        # x/y
        self.ll_extents = extents[:2]  # (x min, y min)
        self.ur_extents = extents[2:]  # (x max, y max)
        self.use_sector_reference = use_sector_reference
        self._transformer = Transformer.from_crs(sector_crs, area_definition.crs)
        super().__init__(area_definition)

    def _get_tile_properties(self, tile_shape, tile_count):
        """Calculate tile information for this particular sector/grid."""
        # ignore tile_shape and tile_count
        # they come from the base class, but aren't used here
        del tile_shape, tile_count

        # get original image's X/Y
        ad = self.area_definition
        x, y = ad.get_proj_vectors()

        ll_xy = self._transformer.transform(*self.ll_extents)
        ur_xy = self._transformer.transform(*self.ur_extents)
        cw = abs(ad.pixel_size_x)
        ch = abs(ad.pixel_size_y)
        st = self.num_subtiles
        cs = self.cell_size  # row height, column width
        # make sure the number of total tiles is a factor of the subtiles
        # meaning each letter has the full number of subtiles
        # Tile numbering/naming starts from the upper left corner
        ul_xy = (ll_xy[0], ur_xy[1])

        # Adjust the upper-left corner to 'perfectly' match the data
        # X/Y are center of pixels, adjust by half a pixels to get upper-left pixel corner
        shift_x = float(ul_xy[0] - (x.min() - cw / 2.)) % cw  # could be negative
        shift_y = float(ul_xy[1] - (y.max() + ch / 2.)) % ch  # could be negative
        # if we're really close to 0 then don't worry about it
        if abs(shift_x) < 1e-10 or abs(shift_x - cw) < 1e-10:
            shift_x = 0
        if abs(shift_y) < 1e-10 or abs(shift_y - ch) < 1e-10:
            shift_y = 0
        if self.use_sector_reference:
            LOG.debug("Adjusting X/Y by (%f, %f) so it better matches lettered grid", shift_x, shift_y)
            x = x + shift_x
            y = y + shift_y
        else:
            LOG.debug("Adjusting lettered grid by (%f, %f) so it better matches data X/Y", shift_x, shift_y)
            ul_xy = (ul_xy[0] - shift_x, ul_xy[1] - shift_y)  # outer edge of grid
            # always keep the same distance between the extents
            ll_xy = (ul_xy[0], ll_xy[1] - shift_y)
            ur_xy = (ur_xy[0] - shift_x, ul_xy[1])

        fcs_y, fcs_x = (np.ceil(float(cs[0]) / st[0]), np.ceil(float(cs[1]) / st[1]))
        # need X/Y for *whole* tiles
        max_cols = np.ceil((ur_xy[0] - ul_xy[0]) / fcs_x)
        max_rows = np.ceil((ul_xy[1] - ll_xy[1]) / fcs_y)
        # don't create partial alpha-tiles
        max_cols = int(np.ceil(max_cols / st[1]) * st[1])
        max_rows = int(np.ceil(max_rows / st[0]) * st[0])

        # make tile cell size a factor of pixel size
        num_pixels_x = int(np.floor(fcs_x / cw))
        num_pixels_y = int(np.floor(fcs_y / ch))
        # NOTE: this does not change the *total* number of columns/rows that
        # will be produced. This is important because otherwise the number
        # of lettered tiles could depend on the input data which is not what we
        # want
        fcs_x = num_pixels_x * cw
        fcs_y = num_pixels_y * ch
        # NOTE: this takes the center of the pixel relative to the upper-left outer edge:
        min_col = max(int(np.floor((x.min() - ul_xy[0]) / fcs_x)), 0)
        max_col = min(int(np.floor((x.max() - ul_xy[0]) / fcs_x)), max_cols - 1)
        min_row = max(int(np.floor((ul_xy[1] - y.max()) / fcs_y)), 0)
        max_row = min(int(np.floor((ul_xy[1] - y.min()) / fcs_y)), max_rows - 1)
        num_cols = max_col - min_col + 1
        num_rows = max_row - min_row + 1

        total_alphas = (max_cols * max_rows) / (st[0] * st[1])
        if total_alphas > 26:
            raise ValueError("Too many lettered grid cells '{}' (sector cell size too small). "
                             "Maximum of 26".format(total_alphas))

        self.tile_shape = (num_pixels_y, num_pixels_x)
        self.total_tile_count = (max_rows, max_cols)
        self.tile_count = (num_rows, num_cols)
        self.total_tiles = num_rows * num_cols
        self.image_shape = (num_pixels_y * num_rows, num_pixels_x * num_cols)
        self.min_col = min_col
        self.max_col = max_col
        self.min_row = min_row
        self.max_row = max_row
        self.ul_xy = ul_xy
        self.mx = cw
        self.bx = ul_xy[0] + cw / 2.0  # X represents the center of the pixel
        self.my = -ch
        self.by = ul_xy[1] - ch / 2.0  # Y represents the center of the pixel
        self.x = x
        self.y = y

    def _get_xy_scaling_parameters(self):
        """Get the X/Y coordinate limits for the full resulting image."""
        return self.mx, self.bx, self.my, self.by

    def _tile_identifier(self, ty, tx):
        """Get tile identifier (name) for a particular tile row/column."""
        st = self.num_subtiles
        ttc = self.total_tile_count
        alpha_num = int((ty // st[0]) * (ttc[1] // st[1]) + (tx // st[1]))
        alpha = string.ascii_uppercase[alpha_num]
        tile_num = int((ty % st[0]) * st[1] + (tx % st[1])) + 1
        return "T{}{:02d}".format(alpha, tile_num)

    def _generate_tile_info(self):
        """Create generator of individual tile metadata."""
        if self._tile_cache:
            for tile_info in self._tile_cache:
                yield tile_info

        ts = self.tile_shape
        ul_xy = self.ul_xy
        x, y = self.x, self.y
        cw = abs(float(self.area_definition.pixel_size_x))
        ch = abs(float(self.area_definition.pixel_size_y))

        # where does the data fall in our lettered grid
        for gy in range(self.min_row, self.max_row + 1):
            for gx in range(self.min_col, self.max_col + 1):
                tile_id = self._tile_identifier(gy, gx)
                # ul_xy is outer-edge of upper-left corner
                # x/y are center of each data pixel
                x_left = ul_xy[0] + gx * ts[1] * cw
                x_right = x_left + ts[1] * cw
                y_top = ul_xy[1] - gy * ts[0] * ch
                y_bot = y_top - ts[0] * ch
                x_mask = np.nonzero((x >= x_left) & (x < x_right))[0]
                y_mask = np.nonzero((y > y_bot) & (y <= y_top))[0]
                if not x_mask.any() or not y_mask.any():
                    # no data in this tile
                    LOG.debug("Tile '%s' doesn't have any data in it", tile_id)
                    continue
                x_slice = slice(x_mask[0], x_mask[-1] + 1)  # assume it's continuous
                y_slice = slice(y_mask[0], y_mask[-1] + 1)

                # theoretically we can precompute the X/Y now
                # instead of taking the x/y data and mapping it
                # to the tile
                tmp_x = np.arange(x_left + cw / 2., x_right, cw)
                tmp_y = np.arange(y_top - ch / 2., y_bot, -ch)
                data_x_idx_min = np.nonzero(np.isclose(tmp_x, x[x_slice.start]))[0][0]
                data_x_idx_max = np.nonzero(np.isclose(tmp_x, x[x_slice.stop - 1]))[0][0]
                # I have a half pixel error some where
                data_y_idx_min = np.nonzero(np.isclose(tmp_y, y[y_slice.start]))[0][0]
                data_y_idx_max = np.nonzero(np.isclose(tmp_y, y[y_slice.stop - 1]))[0][0]
                # now put the data in the grid tile

                tile_slices = (slice(data_y_idx_min, data_y_idx_max + 1),
                               slice(data_x_idx_min, data_x_idx_max + 1))
                data_slices = (y_slice, x_slice)

                tile_number = self._tile_number(gy, gx)
                tile_info = TileInfo(
                    self.tile_count, self.image_shape, ts,
                    gy * ts[0], gx * ts[1], tile_id, tile_number,
                    tmp_x, tmp_y, self.xy_factors, tile_slices, data_slices)
                self._tile_cache.append(tile_info)
                yield tile_info


def _get_factor_offset_fill(input_data_arr, vmin, vmax, encoding):
    dtype_str = encoding["dtype"]
    dtype = np.dtype(getattr(np, dtype_str))
    file_bit_depth = dtype.itemsize * 8
    unsigned_in_signed = encoding.get("_Unsigned") == "true"
    is_unsigned = dtype.kind == "u"
    bit_depth = input_data_arr.attrs.get("bit_depth", file_bit_depth)
    num_fills = 1  # future: possibly support more than one fill value
    if bit_depth is None:
        bit_depth = file_bit_depth
    if bit_depth >= file_bit_depth:
        bit_depth = file_bit_depth
    else:
        # don't take away from the data bit depth if there is room in
        # file data type to allow for extra fill values
        num_fills = 0

    if is_unsigned:
        # max value
        fills = [2 ** file_bit_depth - 1]
    elif unsigned_in_signed:
        # max unsigned value is -1 as a signed int
        fills = [dtype.type(-1)]
    else:
        # max value
        fills = [2 ** (file_bit_depth - 1) - 1]

    # NOTE: AWIPS is buggy and does not properly handle both
    #   halves an integers data space. The below code limits
    #   unsigned integers to the positive half and this seems
    #   to work better with current AWIPS.
    mx = (vmax - vmin) / (2 ** (bit_depth - 1) - 1 - num_fills)
    # NOTE: This is what the line should look like if AWIPS wasn't buggy:
    # mx = (vmax - vmin) / (2 ** bit_depth - 1 - num_fills)
    bx = vmin
    if not is_unsigned and not unsigned_in_signed:
        bx += 2 ** (bit_depth - 1) * mx
    return mx, bx, fills[0]


def _get_data_vmin_vmax(input_data_arr):
    input_metadata = input_data_arr.attrs
    valid_range = input_metadata.get("valid_range")
    if valid_range:
        valid_min, valid_max = valid_range
    else:
        valid_min = input_metadata.get("valid_min")
        valid_max = input_metadata.get("valid_max")
    return valid_min, valid_max


def _add_valid_ranges(data_arrs):
    """Add 'valid_range' metadata if not present.

    If valid_range or valid_min/valid_max are not present in a DataArrays
    metadata (``.attrs``), then lazily compute it with dask so it can be
    computed later when we write tiles out.

    AWIPS requires that scale_factor/add_offset/_FillValue be the **same**
    for all tiles. We must do this calculation before splitting the data into
    tiles otherwise the values will be different.

    """
    for data_arr in data_arrs:
        vmin, vmax = _get_data_vmin_vmax(data_arr)
        if vmin is None:
            # XXX: Do we need to handle category products here?
            vmin = data_arr.min(skipna=True).data
            vmax = data_arr.max(skipna=True).data
            # we don't want to effect the original attrs
            data_arr = data_arr.copy(deep=False)
            # these are dask arrays, they need to get computed later
            data_arr.attrs["valid_range"] = (vmin, vmax)
        yield data_arr


class AWIPSTiledVariableDecisionTree(DecisionTree):
    """Load AWIPS-specific metadata from YAML configuration."""

    def __init__(self, decision_dicts, **kwargs):
        """Initialize decision tree with specific keys to look for."""
        # Fields used to match a product object to it's correct configuration
        attrs = kwargs.pop("attrs",
                           ["name",
                            "standard_name",
                            "satellite",
                            "sensor",
                            "area_id",
                            "units",
                            "reader"]
                           )
        super(AWIPSTiledVariableDecisionTree, self).__init__(decision_dicts, attrs, **kwargs)


class NetCDFTemplate:
    """Helper class to convert a dictionary-based NetCDF template to an :class:`xarray.Dataset`."""

    def __init__(self, template_dict):
        """Parse template dictionary and prepare for rendering."""
        self.is_single_variable = template_dict.get("single_variable", False)
        self.global_attributes = template_dict.get("global_attributes", {})

        default_var_config = {
            "default": {
                "encoding": {"dtype": "uint16"},
            }
        }
        self.variables = template_dict.get("variables", default_var_config)

        default_coord_config = {
            "default": {
                "encoding": {"dtype": "uint16"},
            }
        }
        self.coordinates = template_dict.get("coordinates", default_coord_config)

        self._var_tree = AWIPSTiledVariableDecisionTree([self.variables])
        self._coord_tree = AWIPSTiledVariableDecisionTree([self.coordinates])
        self._filename_format_str = template_dict.get("filename")
        self._str_formatter = StringFormatter()
        self._template_dict = template_dict

    def get_filename(self, base_dir="", **kwargs):
        """Generate output NetCDF file from metadata."""
        # format the filename
        if self._filename_format_str is None:
            raise ValueError("Template does not have a configured "
                             "'filename' pattern.")
        fn_format_str = os.path.join(base_dir, self._filename_format_str)
        filename_parser = Parser(fn_format_str)
        output_filename = filename_parser.compose(kwargs)
        dirname = os.path.dirname(output_filename)
        if dirname and not os.path.isdir(dirname):
            LOG.info("Creating output directory: %s", dirname)
            os.makedirs(dirname)
        return output_filename

    def get_attr_value(self, attr_name, input_metadata, value=None, raw_key=None, raw_value=None, prefix="_"):
        """Determine attribute value using the provided configuration information.

        If `value` and `raw_key` are not provided, this method will search
        for a method named ``<prefix><attr_name>``, which will be called with
        one argument (`input_metadata`) to get the value to return. See
        the documentation for the `prefix` keyword argument below for more
        information.

        Args:
            attr_name (str): Name of the attribute whose value we are
                generating.
            input_metadata (dict): Dictionary of metadata from the input
                DataArray and other context information. Used to provide
                information to `value` or access data from using `raw_key`
                if provided.
            value (Any): Value to assign to this attribute. If a string, it
                may be a python format string which will be provided the data
                from `input_metadata`. For example, ``{name}`` will be filled
                with the value for the ``"name"`` in `input_metadata`. It can
                also include environment variables (ex. ``"${MY_ENV_VAR}"``)
                which will be expanded. String formatting is accomplished by
                the special :class:`trollsift.parser.StringFormatter` which
                allows for special common conversions.
            raw_key (str): Key to access value from `input_metadata`, but
                without any string formatting applied to it. This allows for
                metadata of non-string types to be requested.
            raw_value (Any): Static hardcoded value to set this attribute
                to. Overrides all other options.
            prefix (str): Prefix to use when `value` and `raw_key` are
                both ``None``. Default is ``"_"``. This will be used to find
                custom attribute handlers in subclasses. For example, if
                `value` and `raw_key` are both ``None`` and `attr_name`
                is ``"my_attr"``, then the method ``self._my_attr`` will be
                called as ``return self._my_attr(input_metadata)``.
                See :meth:`NetCDFTemplate.render_global_attributes
                <satpy.writers.awips_tiled.NetCDFTemplate._render_global_attributes>`
                for additional information (prefix is ``"_global_"``).

        """
        if raw_value is not None:
            return raw_value
        if raw_key is not None and raw_key in input_metadata:
            value = input_metadata[raw_key]
            return value

        if isinstance(value, str):
            try:
                value = os.path.expandvars(value)
                value = self._str_formatter.format(value, **input_metadata)
            except (KeyError, ValueError):
                LOG.debug("Can't format string '%s' with provided "
                          "input metadata.", value)
                value = None
                # raise ValueError("Can't format string '{}' with provided "
                #                  "input metadata.".format(value))
        if value is not None:
            return value

        meth_name = prefix + attr_name
        func = getattr(self, meth_name, None)
        if func is not None:
            value = func(input_metadata)
        if value is None:
            LOG.debug("no routine matching %s", meth_name)
        return value

    def _render_attrs(self, attr_configs, input_metadata, prefix="_"):
        attrs = {}
        for attr_name, attr_config_dict in attr_configs.items():
            val = self.get_attr_value(attr_name, input_metadata,
                                      prefix=prefix, **attr_config_dict)
            if val is None:
                # NetCDF attributes can't have a None value
                continue
            attrs[attr_name] = val
        return attrs

    def _render_global_attributes(self, input_metadata):
        attr_configs = self.global_attributes
        return self._render_attrs(attr_configs, input_metadata,
                                  prefix="_global_")

    def _render_variable_attributes(self, var_config, input_metadata):
        attr_configs = var_config["attributes"]
        var_attrs = self._render_attrs(attr_configs, input_metadata, prefix="_data_")
        return var_attrs

    def _render_coordinate_attributes(self, coord_config, input_metadata):
        attr_configs = coord_config["attributes"]
        coord_attrs = self._render_attrs(attr_configs, input_metadata, prefix="_coord_")
        return coord_attrs

    def _render_variable_encoding(self, var_config, input_data_arr):
        new_encoding = input_data_arr.encoding.copy()
        # determine fill value and
        if "encoding" in var_config:
            new_encoding.update(var_config["encoding"])
        if "dtype" not in new_encoding:
            new_encoding["dtype"] = "int16"
            new_encoding["_Unsigned"] = "true"
        return new_encoding

    def _render_variable(self, data_arr):
        var_config = self._var_tree.find_match(**data_arr.attrs)
        new_var_name = var_config.get("var_name", data_arr.attrs["name"])
        new_data_arr = data_arr.copy()
        # remove coords which may cause issues later on
        new_data_arr = new_data_arr.reset_coords(drop=True)

        var_encoding = self._render_variable_encoding(var_config, data_arr)
        new_data_arr.encoding = var_encoding
        var_attrs = self._render_variable_attributes(var_config, data_arr.attrs)
        new_data_arr.attrs = var_attrs
        return new_var_name, new_data_arr

    def _get_matchable_coordinate_metadata(self, coord_name, coord_attrs):
        match_kwargs = {}
        if "name" not in coord_attrs:
            match_kwargs["name"] = coord_name
        match_kwargs.update(coord_attrs)
        return match_kwargs

    def _render_coordinates(self, ds):
        new_coords = {}
        for coord_name, coord_arr in ds.coords.items():
            match_kwargs = self._get_matchable_coordinate_metadata(coord_name, coord_arr.attrs)
            coord_config = self._coord_tree.find_match(**match_kwargs)
            coord_attrs = self._render_coordinate_attributes(coord_config, coord_arr.attrs)
            coord_encoding = self._render_variable_encoding(coord_config, coord_arr)
            new_coords[coord_name] = ds.coords[coord_name].copy()
            new_coords[coord_name].attrs = coord_attrs
            new_coords[coord_name].encoding = coord_encoding
        return new_coords

    def render(
            self,
            dataset_or_data_arrays: xr.Dataset | list[xr.DataArray],
            shared_attrs: dict[str, Any] | None = None,
    ) -> xr.Dataset:
        """Create :class:`xarray.Dataset` from provided data."""
        data_arrays = dataset_or_data_arrays
        if isinstance(data_arrays, xr.Dataset):
            data_arrays = data_arrays.data_vars.values()

        new_ds = xr.Dataset()
        for data_arr in data_arrays:
            new_var_name, new_data_arr = self._render_variable(data_arr)
            new_ds[new_var_name] = new_data_arr
        new_coords = self._render_coordinates(new_ds)
        new_ds.coords.update(new_coords)
        # use first data array as "representative" for global attributes
        # XXX: Should we use global attributes if dataset_or_data_arrays is a Dataset
        if shared_attrs is None:
            shared_attrs = data_arrays[0].attrs
        new_ds.attrs = self._render_global_attributes(shared_attrs)
        return new_ds


class AWIPSNetCDFTemplate(NetCDFTemplate):
    """NetCDF template renderer specifically for tiled AWIPS files."""

    def __init__(self, template_dict, swap_end_time=False):
        """Handle AWIPS special cases and initialize template helpers."""
        self._swap_end_time = swap_end_time
        if swap_end_time:
            self._swap_attributes_end_time(template_dict)
        super().__init__(template_dict)

    def _swap_attributes_end_time(self, template_dict):
        """Swap every use of 'start_time' to use 'end_time' instead."""
        variable_attributes = [var_section["attributes"] for var_section in template_dict.get("variables", {}).values()]
        global_attributes = template_dict.get("global_attributes", {})
        for attr_section in variable_attributes + [global_attributes]:
            for attr_name in attr_section:
                attr_config = attr_section[attr_name]
                if "{start_time" in attr_config.get("value", ""):
                    attr_config["value"] = attr_config["value"].replace("{start_time", "{end_time")
                if attr_config.get("raw_key", "") == "start_time":
                    attr_config["raw_key"] = "end_time"

    def _data_units(self, input_metadata):
        units = input_metadata.get("units", "1")
        # we *know* AWIPS can't handle some units
        return UNIT_CONV.get(units, units)

    def _global_start_date_time(self, input_metadata):
        start_time = input_metadata["start_time"]
        if self._swap_end_time:
            start_time = input_metadata["end_time"]
        return start_time.strftime("%Y-%m-%dT%H:%M:%S")

    def _global_awips_id(self, input_metadata):
        return "AWIPS_" + input_metadata["name"]

    def _global_physical_element(self, input_metadata):
        var_config = self._var_tree.find_match(**input_metadata)
        attr_config = {"physical_element": var_config["attributes"]["physical_element"]}
        result = self._render_attrs(attr_config, input_metadata, prefix="_data_")
        return result["physical_element"]

    def _global_production_location(self, input_metadata):
        """Get default global production_location attribute."""
        del input_metadata
        org = os.environ.get("ORGANIZATION", None)
        if org is not None:
            prod_location = org
        else:
            import socket
            prod_location = socket.gethostname()

        if len(prod_location) > 31:
            warnings.warn(
                "Production location attribute is longer than 31 "
                "characters (AWIPS limit). Set it to a smaller "
                "value with the 'ORGANIZATION' environment "
                "variable. Defaults to hostname and is currently "
                "set to '{}'.".format(prod_location),
                stacklevel=2
            )
            prod_location = prod_location[:31]
        return prod_location

    _global_production_site = _global_production_location

    @staticmethod
    def _get_vmin_vmax(var_config, input_data_arr):
        if "valid_range" in var_config:
            return var_config["valid_range"]
        data_vmin, data_vmax = _get_data_vmin_vmax(input_data_arr)
        return data_vmin, data_vmax

    def _render_variable_encoding(self, var_config, input_data_arr):
        new_encoding = super()._render_variable_encoding(var_config, input_data_arr)
        vmin, vmax = self._get_vmin_vmax(var_config, input_data_arr)
        has_flag_meanings = "flag_meanings" in input_data_arr.attrs
        is_int = np.issubdtype(input_data_arr.dtype, np.integer)
        is_cat = has_flag_meanings or is_int
        has_sf = new_encoding.get("scale_factor") is not None
        if not has_sf and is_cat:
            # AWIPS doesn't like Identity conversion so we can't have
            # a factor of 1 and an offset of 0
            # new_encoding['scale_factor'] = None
            # new_encoding['add_offset'] = None
            if "_FillValue" in input_data_arr.attrs:
                new_encoding["_FillValue"] = input_data_arr.attrs["_FillValue"]
        elif not has_sf and vmin is not None and vmax is not None:
            # calculate scale_factor and add_offset
            sf, ao, fill = _get_factor_offset_fill(
                input_data_arr, vmin, vmax, new_encoding
            )
            # NOTE: These could be dask arrays that will be computed later
            #   when we go to write the files.
            new_encoding["scale_factor"] = sf
            new_encoding["add_offset"] = ao
            new_encoding["_FillValue"] = fill
            new_encoding["coordinates"] = " ".join([ele for ele in input_data_arr.dims])
        return new_encoding

    def _get_projection_attrs(self, area_def):
        """Assign projection attributes per CF standard."""
        proj_attrs = area_def.crs.to_cf()
        proj_encoding = {"dtype": "i4"}
        proj_attrs["short_name"] = area_def.area_id
        gmap_name = proj_attrs["grid_mapping_name"]

        preferred_names = {
            "geostationary": "fixedgrid_projection",
            "lambert_conformal_conic": "lambert_projection",
            "polar_stereographic": "polar_projection",
            "mercator": "mercator_projection",
        }
        if gmap_name not in preferred_names:
            LOG.warning("Data is in projection %s which may not be supported "
                        "by AWIPS", gmap_name)
        area_id_as_var_name = area_def.area_id.replace("-", "_").lower()
        proj_name = preferred_names.get(gmap_name, area_id_as_var_name)
        return proj_name, proj_attrs, proj_encoding

    def _set_xy_coords_attrs(self, new_ds, crs):
        y_attrs = new_ds.coords["y"].attrs
        if crs.is_geographic:
            self._fill_units_and_standard_name(y_attrs, "degrees_north", "latitude")
        else:
            self._fill_units_and_standard_name(y_attrs, "meters", "projection_y_coordinate")
            y_attrs["axis"] = "Y"

        x_attrs = new_ds.coords["x"].attrs
        if crs.is_geographic:
            self._fill_units_and_standard_name(x_attrs, "degrees_east", "longitude")
        else:
            self._fill_units_and_standard_name(x_attrs, "meters", "projection_x_coordinate")
            x_attrs["axis"] = "X"

    @staticmethod
    def _fill_units_and_standard_name(attrs, units, standard_name):
        """Fill in units and standard_name if not set in `attrs`."""
        if attrs.get("units") is None:
            attrs["units"] = units
        if attrs["units"] in ("meter", "metre"):
            # AWIPS doesn't like 'meter'
            attrs["units"] = "meters"
        if attrs.get("standard_name") is None:
            attrs["standard_name"] = standard_name

    def apply_area_def(self, new_ds, area_def):
        """Apply information we can gather from the AreaDefinition."""
        gmap_name, gmap_attrs, gmap_encoding = self._get_projection_attrs(area_def)
        gmap_data_arr = xr.DataArray(0, attrs=gmap_attrs)
        gmap_data_arr.encoding = gmap_encoding
        new_ds[gmap_name] = gmap_data_arr
        self._set_xy_coords_attrs(new_ds, area_def.crs)
        for data_arr in new_ds.data_vars.values():
            if "y" in data_arr.dims and "x" in data_arr.dims:
                data_arr.attrs["grid_mapping"] = gmap_name

        new_ds.attrs["pixel_x_size"] = area_def.pixel_size_x / 1000.0
        new_ds.attrs["pixel_y_size"] = area_def.pixel_size_y / 1000.0
        return new_ds

    def apply_tile_coord_encoding(self, new_ds, xy_factors):
        """Add encoding information specific to the coordinate variables."""
        if "x" in new_ds.coords:
            new_ds.coords["x"].encoding["dtype"] = "int16"
            new_ds.coords["x"].encoding["scale_factor"] = np.float64(xy_factors.mx)
            new_ds.coords["x"].encoding["add_offset"] = np.float64(xy_factors.bx)
        if "y" in new_ds.coords:
            new_ds.coords["y"].encoding["dtype"] = "int16"
            new_ds.coords["y"].encoding["scale_factor"] = np.float64(xy_factors.my)
            new_ds.coords["y"].encoding["add_offset"] = np.float64(xy_factors.by)
        return new_ds

    def apply_tile_info(self, new_ds, tile_info):
        """Apply attributes associated with the current tile."""
        total_tiles = tile_info.tile_count
        total_pixels = tile_info.image_shape
        tile_row = tile_info.tile_row_offset
        tile_column = tile_info.tile_column_offset
        tile_height = new_ds.sizes["y"]
        tile_width = new_ds.sizes["x"]
        new_ds.attrs["tile_row_offset"] = tile_row
        new_ds.attrs["tile_column_offset"] = tile_column
        new_ds.attrs["product_tile_height"] = tile_height
        new_ds.attrs["product_tile_width"] = tile_width
        new_ds.attrs["number_product_tiles"] = total_tiles[0] * total_tiles[1]
        new_ds.attrs["product_rows"] = total_pixels[0]
        new_ds.attrs["product_columns"] = total_pixels[1]
        return new_ds

    def _add_sector_id_global(self, new_ds, sector_id):
        if not self._template_dict.get("add_sector_id_global"):
            return

        if sector_id is None:
            raise ValueError("Keyword 'sector_id' is required for this "
                             "template.")
        new_ds.attrs["sector_id"] = sector_id

    def apply_misc_metadata(self, new_ds, sector_id=None, creator=None, creation_time=None):
        """Add attributes that don't fit into any other category."""
        if creator is None:
            creator = "Satpy Version {} - AWIPS Tiled Writer".format(__version__)
        if creation_time is None:
            creation_time = dt.datetime.now(dt.timezone.utc)

        self._add_sector_id_global(new_ds, sector_id)
        new_ds.attrs["Conventions"] = "CF-1.7"
        new_ds.attrs["creator"] = creator
        new_ds.attrs["creation_time"] = creation_time.strftime("%Y-%m-%dT%H:%M:%S")
        return new_ds

    def _render_variable_attributes(self, var_config, input_metadata):
        attrs = super()._render_variable_attributes(var_config, input_metadata)
        # AWIPS validation checks
        if len(attrs.get("units", "")) > 26:
            warnings.warn(
                "AWIPS 'units' must be limited to a maximum of 26 characters. "
                "Units '{}' is too long and will be truncated.".format(attrs["units"]),
                stacklevel=2
            )
            attrs["units"] = attrs["units"][:26]
        return attrs

    def render(  # type: ignore[override]
            self,
            dataset_or_data_arrays: xr.Dataset | list[xr.DataArray],
            area_def: AreaDefinition,
            tile_info: TileInfo,
            sector_id: str,
            creator: str | None = None,
            creation_time: dt.datetime | None = None,
            shared_attrs: dict[str, Any] | None = None,
            extra_global_attrs: dict[str, Any] | None = None,
    ) -> xr.Dataset:
        """Create a :class:`xarray.Dataset` from template using information provided."""
        new_ds = super().render(dataset_or_data_arrays, shared_attrs=shared_attrs)
        new_ds = self.apply_area_def(new_ds, area_def)
        new_ds = self.apply_tile_coord_encoding(new_ds, tile_info.xy_factors)
        new_ds = self.apply_tile_info(new_ds, tile_info)
        new_ds = self.apply_misc_metadata(new_ds, sector_id, creator, creation_time)
        if extra_global_attrs:
            new_ds.attrs.update(extra_global_attrs)
        return new_ds


def _notnull(data_arr, check_categories=True):
    is_int = np.issubdtype(data_arr.dtype, np.integer)
    fill_value = data_arr.encoding.get("_FillValue", data_arr.attrs.get("_FillValue"))
    if is_int and fill_value is not None:
        # some DQF datasets are always valid
        if check_categories:
            return data_arr != fill_value
        return False
    return data_arr.notnull()


def _any_notnull(data_arr, check_categories):
    not_null = _notnull(data_arr, check_categories)
    if not_null is False:
        return False
    return not_null.any()


def _is_empty_tile(dataset_to_save, check_categories):
    # check if this tile is empty
    # if so, don't create it
    for data_var in dataset_to_save.data_vars.values():
        if data_var.ndim == 2 and _any_notnull(data_var, check_categories):
            return False
    return True


def _copy_to_existing(dataset_to_save, output_filename):
    # Experimental: This function doesn't seem to behave well with xarray file
    #   caching and/or multiple dask workers. It causes tests to hang, but
    #   only sometimes. Limiting dask to 1 worker seems to fix this.
    #   I (David Hoese) was unable to make a script that reproduces this
    #   without using this writer (makes it difficult to file a bug report).
    existing_dataset = xr.open_dataset(output_filename, cache=False)
    # the below used to trick xarray into working, but this doesn't work
    # in newer versions. This was a hack in the first place so I'm keeping it
    # here for reference.
    # existing_dataset = existing_dataset.copy(deep=True)
    # existing_dataset.close()

    # update existing data with new valid data
    for var_name, var_data_arr in dataset_to_save.data_vars.items():
        if var_name not in existing_dataset:
            continue
        if var_data_arr.ndim != 2:
            continue
        existing_data_arr = existing_dataset[var_name]
        valid_current = _notnull(var_data_arr)
        new_data = existing_data_arr.data[:]
        new_data[valid_current] = var_data_arr.data[valid_current]
        var_data_arr.data[:] = new_data
        var_data_arr.encoding.update(existing_data_arr.encoding)
        var_data_arr.encoding.pop("source", None)

    return dataset_to_save


def _extract_factors(dataset_to_save):
    factors = {}
    for data_var in dataset_to_save.data_vars.values():
        enc = data_var.encoding
        data_var.attrs.pop("valid_range", None)
        factor_set = (enc.pop("scale_factor", None),
                      enc.pop("add_offset", None),
                      enc.pop("_FillValue", None))
        factors[data_var.name] = factor_set
    return factors


def _reapply_factors(dataset_to_save, factors):
    for var_name, factor_set in factors.items():
        data_arr = dataset_to_save[var_name]
        if factor_set[0] is not None:
            data_arr.encoding["scale_factor"] = factor_set[0]
        if factor_set[1] is not None:
            data_arr.encoding["add_offset"] = factor_set[1]
        if factor_set[2] is not None:
            data_arr.encoding["_FillValue"] = factor_set[2]
    return dataset_to_save


def to_nonempty_netcdf(dataset_to_save: xr.Dataset,
                       output_filename: str,
                       update_existing: bool = True,
                       check_categories: bool = True) -> None:
    """Save :class:`xarray.Dataset` to a NetCDF file if not all fills.

    In addition to checking certain Dataset variables for fill values,
    this function can also "update" an existing NetCDF file with the
    new valid data provided.

    """
    if _is_empty_tile(dataset_to_save, check_categories):
        LOG.debug("Skipping tile creation for %s because it would be "
                  "empty.", output_filename)
        return

    # TODO: Allow for new variables to be created
    if update_existing and os.path.isfile(output_filename):
        dataset_to_save = _copy_to_existing(dataset_to_save, output_filename)
        mode = "a"
    else:
        mode = "w"
    with warnings.catch_warnings():
        # this is an expected warning as CF convention tells us not to have a _FillValue for coordinate variables
        warnings.filterwarnings(
            "ignore",
            message="saving variable [xy] with .* without any _FillValue.*",
            category=xr.SerializationWarning,
        )
        LOG.info(f"Saving AWIPS tile {output_filename}...")
        dataset_to_save.to_netcdf(output_filename, mode=mode)


def tile_filler(data_arr_data, tile_shape, tile_slices, fill_value):
    """Create an empty tile array and fill the proper locations with data."""
    empty_tile = np.full(tile_shape, fill_value, dtype=data_arr_data.dtype)
    empty_tile[tile_slices] = data_arr_data
    return empty_tile


class AWIPSTiledWriter(Writer):
    """Writer for AWIPS NetCDF4 Tile files.

    See :mod:`satpy.writers.awips_tiled` documentation for more information
    on templates and produced file format.

    """

    def __init__(self, compress=False, **kwargs):
        """Initialize writer and decision trees."""
        super(AWIPSTiledWriter, self).__init__(default_config_filename="writers/awips_tiled.yaml", **kwargs)
        self.base_dir = kwargs.get("base_dir", "")
        self.awips_sectors = self.config["sectors"]
        self.templates = self.config["templates"]
        self.compress = compress
        self._fill_sector_info()
        self._enhancer = None

    @property
    def enhancer(self):
        """Get lazy loaded enhancer object only if needed."""
        if self._enhancer is None:
            from satpy.enhancements.enhancer import Enhancer

            self._enhancer = Enhancer()
        return self._enhancer

    @classmethod
    def separate_init_kwargs(cls, kwargs):
        """Separate keyword arguments by initialization and saving keyword arguments."""
        # FUTURE: Don't pass Scene.save_datasets kwargs to init and here
        init_kwargs, kwargs = super(AWIPSTiledWriter, cls).separate_init_kwargs(
            kwargs)
        for kw in ["compress"]:
            if kw in kwargs:
                init_kwargs[kw] = kwargs.pop(kw)

        return init_kwargs, kwargs

    def _fill_sector_info(self):
        """Convert sector extents if needed."""
        for sector_info in self.awips_sectors.values():
            sector_info["projection"] = CRS.from_user_input(sector_info["projection"])
            p = Proj(sector_info["projection"])
            if "lower_left_xy" in sector_info:
                sector_info["lower_left_lonlat"] = p(*sector_info["lower_left_xy"], inverse=True)
            else:
                sector_info["lower_left_xy"] = p(*sector_info["lower_left_lonlat"])
            if "upper_right_xy" in sector_info:
                sector_info["upper_right_lonlat"] = p(*sector_info["upper_right_xy"], inverse=True)
            else:
                sector_info["upper_right_xy"] = p(*sector_info["upper_right_lonlat"])

    def _get_lettered_sector_info(self, sector_id):
        """Get metadata for the current sector if configured.

        This is not necessary for numbered grids. If found, the sector info
        will provide the overall tile layout for this grid/sector. This allows
        for consistent tile numbering/naming regardless of where the data being
        converted actually is.

        """
        if sector_id is None:
            raise TypeError("Keyword 'sector_id' is required for lettered grids.")
        try:
            return self.awips_sectors[sector_id]
        except KeyError:
            raise ValueError("Unknown sector '{}'".format(sector_id))

    def _get_tile_generator(self, area_def, lettered_grid, sector_id,
                            num_subtiles, tile_size, tile_count,
                            use_sector_reference=False):
        """Get the appropriate tile generator class for lettered or numbered tiles."""
        # Create a tile generator for this grid definition
        if lettered_grid:
            sector_info = self._get_lettered_sector_info(sector_id)
            tile_gen = LetteredTileGenerator(
                area_def,
                sector_info["lower_left_xy"] + sector_info["upper_right_xy"],
                sector_crs=sector_info["projection"],
                cell_size=sector_info["resolution"],
                num_subtiles=num_subtiles,
                use_sector_reference=use_sector_reference,
                )
        else:
            tile_gen = NumberedTileGenerator(
                area_def,
                tile_shape=tile_size,
                tile_count=tile_count,
            )
        return tile_gen

    def _group_by_area(self, datasets):
        """Group datasets by their area."""
        def _area_id(area_def):
            return area_def.description + str(area_def.area_extent) + str(area_def.shape)

        # get all of the datasets stored by area
        area_datasets = {}
        for x in datasets:
            area_id = _area_id(x.attrs["area"])
            area, ds_list = area_datasets.setdefault(area_id, (x.attrs["area"], []))
            ds_list.append(x)
        return area_datasets

    def _split_rgbs(self, ds):
        """Split a single RGB dataset in to multiple."""
        for component in "RGB":
            band_data = ds.sel(bands=component)
            band_data.attrs["name"] += "_{}".format(component)
            band_data.attrs["valid_min"] = 0.0
            band_data.attrs["valid_max"] = 1.0
            yield band_data

    def _enhance_and_split_rgbs(self, datasets):
        """Handle multi-band images by splitting in to separate products."""
        new_datasets = []
        for ds in datasets:
            if ds.ndim == 2:
                new_datasets.append(ds)
                continue
            elif ds.ndim > 3 or ds.ndim < 1 or (ds.ndim == 3 and "bands" not in ds.coords):
                LOG.error("Can't save datasets with more or less than 2 dimensions "
                          "that aren't RGBs to AWIPS Tiled format: %s", ds.name)
            else:
                # this is an RGB
                from satpy.enhancements.enhancer import get_enhanced_image

                img = get_enhanced_image(ds.squeeze(), enhance=self.enhancer)
                res_data = img.finalize(fill_value=0, dtype=np.float32)[0]
                new_datasets.extend(self._split_rgbs(res_data))

        return new_datasets

    def _tile_filler(self, tile_info, data_arr):
        fill = np.nan if np.issubdtype(data_arr.dtype, np.floating) else data_arr.attrs.get("_FillValue", 0)
        data_arr_data = data_arr.data[tile_info.data_slices]
        data_arr_data = data_arr_data.rechunk(data_arr_data.shape)
        # NOTE: Later steps assume this step produces tiles as a single dask chunks
        new_data = da.map_blocks(tile_filler, data_arr_data,
                                 tile_info.tile_shape, tile_info.tile_slices,
                                 fill, dtype=data_arr.dtype, chunks=tile_info.tile_shape)
        return xr.DataArray(new_data, dims=("y", "x"),
                            attrs=data_arr.attrs.copy())

    def _slice_and_update_coords(self, tile_info, data_arrays):
        new_x = xr.DataArray(tile_info.x, dims=("x",))
        if "x" in data_arrays[0].coords:
            old_x = data_arrays[0].coords["x"]
            new_x.attrs.update(old_x.attrs)
            new_x.encoding = old_x.encoding
        new_y = xr.DataArray(tile_info.y, dims=("y",))
        if "y" in data_arrays[0].coords:
            old_y = data_arrays[0].coords["y"]
            new_y.attrs.update(old_y.attrs)
            new_y.encoding = old_y.encoding

        for data_arr in data_arrays:
            new_data_arr = self._tile_filler(tile_info, data_arr)
            new_data_arr.coords["x"] = new_x
            new_data_arr.coords["y"] = new_y
            yield new_data_arr

    def _iter_tile_info_and_datasets(self, tile_gen, data_arrays, single_variable=True):
        all_data_arrays = self._enhance_and_split_rgbs(data_arrays)
        if single_variable:
            all_data_arrays = [[single_data_arr] for single_data_arr in all_data_arrays]
        else:
            all_data_arrays = [all_data_arrays]
        for data_arrays_set in all_data_arrays:
            for tile_info in tile_gen():
                data_arrays_tile_set = list(self._slice_and_update_coords(tile_info, data_arrays_set))
                yield tile_info, data_arrays_tile_set

    def _iter_area_tile_info_and_datasets(self, area_datasets, template,
                                          lettered_grid, sector_id,
                                          num_subtiles, tile_size, tile_count,
                                          use_sector_reference):
        for area_def, data_arrays in area_datasets.values():
            data_arrays = list(_add_valid_ranges(data_arrays))
            tile_gen = self._get_tile_generator(
                area_def, lettered_grid, sector_id, num_subtiles, tile_size,
                tile_count, use_sector_reference=use_sector_reference)
            for tile_info, data_arrs in self._iter_tile_info_and_datasets(
                    tile_gen, data_arrays, single_variable=template.is_single_variable):
                yield area_def, tile_info, data_arrs

    def save_dataset(self, dataset, **kwargs):
        """Save a single DataArray to one or more NetCDF4 Tile files."""
        LOG.warning("For best performance use `save_datasets`")
        return self.save_datasets([dataset], **kwargs)

    def get_filename(self, template, area_def, tile_info, sector_id, **kwargs):
        """Generate output NetCDF file from metadata."""
        # format the filename
        try:
            return super(AWIPSTiledWriter, self).get_filename(
                area_id=area_def.area_id,
                rows=area_def.height,
                columns=area_def.width,
                sector_id=sector_id,
                tile_id=tile_info.tile_id,
                tile_number=tile_info.tile_number,
                **kwargs)
        except RuntimeError:
            # the user didn't provide a specific filename, use the template
            return template.get_filename(
                base_dir=self.base_dir,
                area_id=area_def.area_id,
                rows=area_def.height,
                columns=area_def.width,
                sector_id=sector_id,
                tile_id=tile_info.tile_id,
                tile_number=tile_info.tile_number,
                **kwargs)

    def check_tile_exists(self, output_filename):
        """Check if tile exists and report error accordingly."""
        if os.path.isfile(output_filename):
            LOG.info("AWIPS file already exists, will update with new data: %s", output_filename)

    def _adjust_metadata_times(self, ds_info):
        debug_shift_time = int(os.environ.get("DEBUG_TIME_SHIFT", 0))
        if debug_shift_time:
            ds_info["start_time"] += dt.timedelta(minutes=debug_shift_time)
            ds_info["end_time"] += dt.timedelta(minutes=debug_shift_time)

    def _get_tile_data_info(self, data_arrs, creation_time, source_name):
        # use the first data array as a "representative" for the group
        ds_info = data_arrs[0].attrs.copy()
        del ds_info["valid_range"]  # remove variable-specific metadata that may contain dask arrays
        # we want to use our own creation_time
        ds_info["creation_time"] = creation_time
        if source_name is not None:
            ds_info["source_name"] = source_name
        self._adjust_metadata_times(ds_info)
        return ds_info

    # TODO: Add additional untiled variable support
    def save_datasets(self, datasets, sector_id=None,  # noqa: D417
                      source_name=None,
                      tile_count=(1, 1), tile_size=None,
                      lettered_grid=False, num_subtiles=None,
                      use_end_time=False, use_sector_reference=False,
                      template="polar", check_categories=True,
                      extra_global_attrs=None, environment_prefix="DR",
                      compute=True, **kwargs):
        """Write a series of DataArray objects to multiple NetCDF4 Tile files.

        Args:
            datasets (Iterable): Series of gridded :class:`~xarray.DataArray`
                objects with the necessary metadata to be converted to a valid
                tile product file.
            sector_id (str, Optional): Name of the region or sector that the provided
                data is on. This name will be written to the NetCDF file and
                will be used as the sector in the AWIPS client for the 'polar'
                template. For lettered
                grids this name should match the name configured in the writer
                YAML. This is required for some templates (ex. default 'polar'
                template) but is defined as a keyword argument
                for better error handling in Satpy.
            source_name (str, Optional): Name of producer of these files (ex. "SSEC").
                This name is used to create the output filename for some
                templates.
            environment_prefix (str, Optional): Prefix of filenames for some templates.
                For operational real-time data this is usually "OR", "OT" for
                test data, "IR" for test system real-time data, and "IT" for
                test system test data. This defaults to "DR" for "Developer
                Real-time" to avoid anyone accidentally producing files that
                could be mistaken for the operational system.
            tile_count (tuple, Optional): For numbered tiles only, how many tile rows
                and tile columns to produce. Default to ``(1, 1)``, a single
                giant tile. Either ``tile_count``, ``tile_size``, or
                ``lettered_grid`` should be specified.
            tile_size (tuple, Optional): For numbered tiles only, how many pixels each
                tile should be. This takes precedence over ``tile_count`` if
                specified. Either ``tile_count``, ``tile_size``, or
                ``lettered_grid`` should be specified.
            lettered_grid (bool, Optional): Whether to use a preconfigured grid and
                label tiles with letters and numbers instead of only numbers.
                For example, tiles will be named "A01", "A02", "B01", and so
                on in the first row of data and continue on to "A03", "A04",
                and "B03" in the default case where ``num_subtiles`` is (2, 2).
                Letters start in the upper-left corner and will go from A up to
                Z, if necessary.
            num_subtiles (tuple, Optional): For lettered tiles only, how many rows and
                columns to split each lettered tile in to. By default 2 rows
                and 2 columns will be created. For example, the tile for
                letter "A" will have "A01" and "A02" in the top row and "A03"
                and "A04" in the second row.
            use_end_time (bool, Optional): Instead of using the ``start_time`` for the
                product filename and time written to the file, use the
                ``end_time``. This is useful for multi-day composites where
                the ``end_time`` is a better representation of what data is
                in the file.
            use_sector_reference (bool, Optional): For lettered tiles only, whether to
                shift the data locations to align with the preconfigured
                grid's pixels. By default this is False meaning that the
                grid's tiles will be shifted to align with the data locations.
                If True, the data is shifted. At most the data will be shifted
                by 0.5 pixels. See :mod:`satpy.writers.awips_tiled` for more
                information.
            template (str or dict, Optional): Name of the template configured in the
                writer YAML file. This can also be a dictionary with a full
                template configuration. See the :mod:`satpy.writers.awips_tiled`
                documentation for more information on templates. Defaults to
                the 'polar' builtin template.
            check_categories (bool, Optional): Whether category and flag products should
                be included in the checks for empty or not empty tiles. In
                some cases (ex. data quality flags) category products may look
                like all valid data (a non-empty tile) but shouldn't be used
                to determine the emptiness of the overall tile (good quality
                versus non-existent). Default is True. Set to False to ignore
                category (integer dtype or "flag_meanings" defined) when
                checking for valid data.
            extra_global_attrs (dict, Optional): Additional global attributes to be
                added to every produced file. These attributes are applied
                at the end of template rendering and will therefore overwrite
                template generated values with the same global attribute name.
            compute (bool, Optional): Compute and write the output immediately using
                dask. Default to ``False``.

        """
        if not isinstance(template, dict):
            template = self.config["templates"][template]
        template = AWIPSNetCDFTemplate(template, swap_end_time=use_end_time)
        area_data_arrs = self._group_by_area(datasets)

        arrays_to_compute = []
        creation_time = dt.datetime.now(dt.timezone.utc)
        area_tile_data_gen = self._iter_area_tile_info_and_datasets(
            area_data_arrs, template, lettered_grid, sector_id, num_subtiles,
            tile_size, tile_count, use_sector_reference)
        for area_def, tile_info, data_arrs in area_tile_data_gen:
            # TODO: Create Dataset object of all of the sliced-DataArrays (optional)
            ds_info = self._get_tile_data_info(data_arrs,
                                               creation_time,
                                               source_name)
            output_filename = self.get_filename(template, area_def,
                                                tile_info, sector_id,
                                                environment_prefix=environment_prefix,
                                                **ds_info)
            self.check_tile_exists(output_filename)

            render_kwargs = {
                "area_def": area_def,
                "tile_info": tile_info,
                "sector_id": sector_id,
                "creation_time": creation_time,
                "shared_attrs": ds_info,
                "extra_global_attrs": extra_global_attrs,
            }
            res = _save_tile_data_arrays(
                data_arrs,
                output_filename,
                template,
                self.compress,
                check_categories,
                render_kwargs,
            )

            arrays_to_compute.append(res)
        if not arrays_to_compute:
            # no tiles produced
            return []

        if not compute:
            return arrays_to_compute
        return dask.compute(arrays_to_compute)


def _save_tile_data_arrays(
        data_arrs: list[xr.DataArray],
        output_filename: str,
        template: NetCDFTemplate,
        compress: bool,
        check_categories: bool,
        render_kwargs: dict[str, Any],
) -> da.Array:
    data_arr_dims_pairs = tuple(
        elem for data_arr in data_arrs
        for elem in
        (
            data_arr.data,
            # mypy complains because ".dims" are defined as Hashable by xarray, not strings
            "".join(str(dim) for dim in data_arr.dims),
        )
    )
    # Convert xarray's internal dict-like objects to pure dicts so dask
    # tokenizing doesn't try generic object tokenization
    all_attrs = [dict(data_arr.attrs) for data_arr in data_arrs]
    all_coords = [dict(data_arr.coords) for data_arr in data_arrs]
    res = da.blockwise(
        _save_tile_block,
        "a",
        *data_arr_dims_pairs,
        new_axes={"a": 1},
        meta=np.ndarray((), dtype=object),
        dtype=object,
        all_attrs=all_attrs,
        all_coords=all_coords,
        template=template,
        compress=compress,
        check_categories=check_categories,
        output_filename=output_filename,
        render_kwargs=render_kwargs,
    )
    return res


def _save_tile_block(
        *input_arrays: xr.DataArray,
        output_filename: str,
        template: NetCDFTemplate,
        compress: bool,
        check_categories: bool,
        render_kwargs: dict[str, Any],
        all_attrs: list[dict],
        all_coords: list[dict],
) -> str:
    restruct_data_arrs = [
        xr.DataArray(np_arr_list[0][0], attrs=all_attrs[idx], dims=("y", "x"), coords=all_coords[idx])
        for idx, np_arr_list in enumerate(input_arrays)
    ]
    # TODO: Provide attribute caching for things that likely won't change (functools lrucache)
    new_ds = template.render(restruct_data_arrs, **render_kwargs)
    if compress:
        new_ds.encoding["zlib"] = True
        for var in new_ds.variables.values():
            var.encoding["zlib"] = True
    to_nonempty_netcdf(new_ds, output_filename, update_existing=True, check_categories=check_categories)
    return output_filename


def _create_debug_array(sector_info, num_subtiles, font_path="Verdana.ttf"):
    from PIL import Image, ImageDraw, ImageFont
    from pkg_resources import resource_filename as get_resource_filename
    size = (1000, 1000)
    img = Image.new("L", size, 0)
    draw = ImageDraw.Draw(img)

    if ":" in font_path:
        # load from a python package
        font_path = get_resource_filename(*font_path.split(":"))
    font = ImageFont.truetype(font_path, 25)

    ll_extent = sector_info["lower_left_xy"]
    ur_extent = sector_info["upper_right_xy"]
    total_meters_x = ur_extent[0] - ll_extent[0]
    total_meters_y = ur_extent[1] - ll_extent[1]
    fcs_x = np.ceil(float(sector_info["resolution"][1]) / num_subtiles[1])
    fcs_y = np.ceil(float(sector_info["resolution"][0]) / num_subtiles[0])
    total_cells_x = np.ceil(total_meters_x / fcs_x)
    total_cells_y = np.ceil(total_meters_y / fcs_y)
    total_cells_x = np.ceil(total_cells_x / num_subtiles[1]) * num_subtiles[1]
    total_cells_y = np.ceil(total_cells_y / num_subtiles[0]) * num_subtiles[0]
    # total_alpha_cells_x = int(total_cells_x / num_subtiles[1])
    # total_alpha_cells_y = int(total_cells_y / num_subtiles[0])

    # "round" the total meters up to the number of alpha cells
    # total_meters_x = total_cells_x * fcs_x
    # total_meters_y = total_cells_y * fcs_y

    # Pixels per tile
    ppt_x = np.floor(float(size[0]) / total_cells_x)
    ppt_y = np.floor(float(size[1]) / total_cells_y)
    half_ppt_x = np.floor(ppt_x / 2.)
    half_ppt_y = np.floor(ppt_y / 2.)
    # Meters per pixel
    meters_ppx = fcs_x / ppt_x
    meters_ppy = fcs_y / ppt_y
    for idx, alpha in enumerate(string.ascii_uppercase):
        for i in range(4):
            st_x = i % num_subtiles[1]
            st_y = int(i / num_subtiles[1])
            t = "{}{:02d}".format(alpha, i + 1)
            t_size = font.getsize(t)
            cell_x = (idx * num_subtiles[1] + st_x) % total_cells_x
            cell_y = int(idx / (total_cells_x / num_subtiles[1])) * num_subtiles[0] + st_y
            if (cell_x > total_cells_x) or (cell_y > total_cells_y):
                continue
            x = ppt_x * cell_x + half_ppt_x
            y = ppt_y * cell_y + half_ppt_y
            # draw box around the tile edge
            # PIL Documentation: "The second point is just outside the drawn rectangle."
            # we want to be just inside 0 and just inside the outer edge of the tile
            draw_rectangle(draw,
                           (x - half_ppt_x, y - half_ppt_y,
                            x + half_ppt_x, y + half_ppt_y), outline=255, fill=75, width=3)
            draw.text((x - t_size[0] / 2., y - t_size[1] / 2.), t, fill=255, font=font)

    img.save("test.png")

    new_extents = (
        ll_extent[0],
        ur_extent[1] - 1001. * meters_ppy,
        ll_extent[0] + 1001. * meters_ppx,
        ur_extent[1],
    )
    grid_def = AreaDefinition(
        "debug_grid",
        "debug_grid",
        "debug_grid",
        sector_info["projection"],
        1000,
        1000,
        new_extents
    )
    return grid_def, np.array(img)


def draw_rectangle(draw, coordinates, outline=None, fill=None, width=1):
    """Draw simple rectangle in to a numpy array image."""
    for i in range(width):
        rect_start = (coordinates[0] + i, coordinates[1] + i)
        rect_end = (coordinates[2] - i, coordinates[3] - i)
        draw.rectangle((rect_start, rect_end), outline=outline, fill=fill)


def create_debug_lettered_tiles(**writer_kwargs):
    """Create tile files with tile identifiers "burned" in to the image data for debugging."""
    writer_kwargs["lettered_grid"] = True
    writer_kwargs["num_subtiles"] = (2, 2)  # default, don't use command line argument

    init_kwargs, save_kwargs = AWIPSTiledWriter.separate_init_kwargs(**writer_kwargs)
    writer = AWIPSTiledWriter(**init_kwargs)

    sector_id = save_kwargs["sector_id"]
    sector_info = writer.awips_sectors[sector_id]
    area_def, arr = _create_debug_array(sector_info, save_kwargs["num_subtiles"])

    now = dt.datetime.now(dt.timezone.utc)
    product = xr.DataArray(da.from_array(arr, chunks="auto"), attrs=dict(
        name="debug_{}".format(sector_id),
        platform_name="DEBUG",
        sensor="TILES",
        start_time=now,
        end_time=now,
        area=area_def,
        standard_name="toa_bidirectional_reflectance",
        units="1",
        valid_min=0,
        valid_max=255,
    ))
    created_files = writer.save_dataset(
        product,
        **save_kwargs
    )
    return created_files


def main():
    """Command line interface mimicing CSPP Polar2Grid."""
    import argparse
    parser = argparse.ArgumentParser(description="Create AWIPS compatible NetCDF tile files")
    parser.add_argument("--create-debug", action="store_true",
                        help="Create debug NetCDF files to show tile locations in AWIPS")
    parser.add_argument("-v", "--verbose", dest="verbosity", action="count", default=0,
                        help="each occurrence increases verbosity 1 level through "
                             "ERROR-WARNING-INFO-DEBUG (default INFO)")
    parser.add_argument("-l", "--log", dest="log_fn", default=None,
                        help="specify the log filename")

    group_1 = parser.add_argument_group(title="Writer Initialization")
    group_1.add_argument("--backend-configs", nargs="*", dest="backend_configs",
                         help="alternative backend configuration files")
    group_1.add_argument("--compress", action="store_true",
                         help="zlib compress each netcdf file")
    # group_1.add_argument("--fix-awips", action="store_true",
    #                      help="modify NetCDF output to work with the old/broken AWIPS NetCDF library")
    group_2 = parser.add_argument_group(title="Wrtier Save")
    group_2.add_argument("--tiles", dest="tile_count", nargs=2, type=int, default=[1, 1],
                         help="Number of tiles to produce in Y (rows) and X (cols) direction respectively")
    group_2.add_argument("--tile-size", dest="tile_size", nargs=2, type=int, default=None,
                         help="Specify how many pixels are in each tile (overrides '--tiles')")
    # group.add_argument('--tile-offset', nargs=2, default=(0, 0),
    #                    help="Start counting tiles from this offset ('row_offset col_offset')")
    group_2.add_argument("--letters", dest="lettered_grid", action="store_true",
                         help="Create tiles from a static letter-based grid based on the product projection")
    group_2.add_argument("--letter-subtiles", nargs=2, type=int, default=(2, 2),
                         help="Specify number of subtiles in each lettered tile: 'row col'")
    group_2.add_argument("--output-pattern", default=DEFAULT_OUTPUT_PATTERN,
                         help="output filenaming pattern")
    group_2.add_argument("--source-name", default="SSEC",
                         help="specify processing source name used in attributes and filename (default 'SSEC')")
    group_2.add_argument("--sector-id", required=True,
                         help="specify name for sector/region used in attributes and filename (example 'LCC')")
    group_2.add_argument("--template", default="polar",
                         help="specify the template name to use (default: polar)")
    args = parser.parse_args()

    # Logs are renamed once data the provided start date is known
    levels = [logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG]
    logging.basicConfig(level=levels[min(3, args.verbosity)], filename=args.log_fn)

    if args.create_debug:
        writer_kwargs = vars(args)
        create_debug_lettered_tiles(**writer_kwargs)
        return
    else:
        raise NotImplementedError("Command line interface not implemented yet for AWIPS tiled writer")


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2015-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Generic PIL/Pillow image format writer."""
from __future__ import annotations

import logging
import typing

from satpy.writers.core.image import ImageWriter

if typing.TYPE_CHECKING:
    from os import PathLike
    from typing import Any

    import dask.array as da
    from dask.delayed import Delayed
    from trollimage.xrimage import XRImage

LOG = logging.getLogger(__name__)


class PillowWriter(ImageWriter):
    """Generic PIL image format writer."""

    def __init__(self, **kwargs):
        """Initialize image writer plugin."""
        ImageWriter.__init__(
            self,
            default_config_filename="writers/simple_image.yaml",
            **kwargs)

    def save_image(
        self,
        img: XRImage,
        filename: str | None = None,
        compute: bool = True,
        **kwargs,
    ) -> list[da.Array | Delayed] | tuple[list[da.Array], list[Any]] | list[str | PathLike | None]:
        """Save Image object to a given ``filename``.

        Args:
            img: Image object to save to disk.
            filename: Optionally specify the filename to save this
                dataset to. It may include string formatting
                patterns that will be filled in by dataset
                attributes.
            compute: If `True` (default), compute and save the dataset.
                If `False` return either a `dask.delayed.Delayed`
                object or tuple of (source, target). See the
                return values below for more information.
            **kwargs: Keyword arguments to pass to the images `save` method.

        Returns:
            Value returned depends on `compute`. If `compute` is `True` then
            the return value is the result of computing a
            `dask.delayed.Delayed` object or running `dask.array.store`. If
            `compute` is `False` then the returned value is either a
            `dask.delayed.Delayed` object that can be computed using
            `delayed.compute()` or a tuple of (source, target) that should be
            passed to `dask.array.store`. If target is provided the the caller
            is responsible for calling `target.close()` if the target has
            this method.

        """
        filename = filename or self.get_filename(**img.data.attrs)

        LOG.debug("Saving to image: %s", filename)
        res = img.save(filename, compute=compute, **kwargs)
        # old trollimage <1.27.0: res is None or Delayed
        # new trollimage: res is str | Path or dask Array
        return [res]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""The hrit msg reader tests package."""

import datetime
import unittest
from unittest import mock

import numpy as np
from xarray import DataArray

from satpy.readers.goes_imager_hrit import (
    ALTITUDE,
    HRITGOESFileHandler,
    HRITGOESPrologueFileHandler,
    make_gvar_float,
    make_sgs_time,
    sgs_time,
)
from satpy.tests.utils import make_dataid


class TestGVARFloat(unittest.TestCase):
    """GVAR float tester."""

    def test_fun(self):
        """Test function."""
        test_data = [(-1.0, b"\xbe\xf0\x00\x00"),
                     (-0.1640625, b"\xbf\xd6\x00\x00"),
                     (0.0, b"\x00\x00\x00\x00"),
                     (0.1640625, b"\x40\x2a\x00\x00"),
                     (1.0, b"\x41\x10\x00\x00"),
                     (100.1640625, b"\x42\x64\x2a\x00")]

        for expected, str_val in test_data:
            val = np.frombuffer(str_val, dtype=">i4")
            assert expected == make_gvar_float(val)


class TestMakeSGSTime(unittest.TestCase):
    """SGS Time tester."""

    def test_fun(self):
        """Encode the test time."""
        # 2018-129 (may 9th), 21:33:27.999
        tcds = np.array([(32, 24, 18, 146, 19, 50, 121, 153)], dtype=sgs_time)
        expected = datetime.datetime(2018, 5, 9, 21, 33, 27, 999000)
        assert make_sgs_time(tcds[0]) == expected


test_pro = {"TISTR": datetime.datetime(2018, 5, 9, 21, 33, 27, 999000),
            "TCurr": datetime.datetime(2018, 5, 9, 21, 33, 27, 999000),
            "TCLMT": datetime.datetime(2018, 5, 9, 21, 33, 27, 999000),
            "SubSatLongitude": 100.1640625,
            "TCHED": datetime.datetime(2018, 5, 9, 21, 33, 27, 999000),
            "TLTRL": datetime.datetime(2018, 5, 9, 21, 33, 27, 999000),
            "TIPFS": datetime.datetime(2018, 5, 9, 21, 33, 27, 999000),
            "TISPC": datetime.datetime(2018, 5, 9, 21, 33, 27, 999000),
            "ReferenceLatitude": 0.0,
            "TIIRT": datetime.datetime(2018, 5, 9, 21, 33, 27, 999000),
            "TLHED": datetime.datetime(2018, 5, 9, 21, 33, 27, 999000),
            "TIVIT": datetime.datetime(2018, 5, 9, 21, 33, 27, 999000),
            "SubSatLatitude": 0.0,
            "TIECL": datetime.datetime(2018, 5, 9, 21, 33, 27, 999000),
            "ReferenceLongitude": 100.1640625,
            "TCTRL": datetime.datetime(2018, 5, 9, 21, 33, 27, 999000),
            "TLRAN": datetime.datetime(2018, 5, 9, 21, 33, 27, 999000),
            "TINFS": datetime.datetime(2018, 5, 9, 21, 33, 27, 999000),
            "TIBBC": datetime.datetime(2018, 5, 9, 21, 33, 27, 999000),
            "TIONA": datetime.datetime(2018, 5, 9, 21, 33, 27, 999000),
            "ReferenceDistance": 100.1640625,
            "SatelliteID": 15}


class TestHRITGOESPrologueFileHandler(unittest.TestCase):
    """Test the HRITFileHandler."""

    @mock.patch("satpy.readers.goes_imager_hrit.recarray2dict")
    @mock.patch("satpy.readers.goes_imager_hrit.np.fromfile")
    @mock.patch("satpy.readers.goes_imager_hrit.HRITFileHandler.__init__")
    def test_init(self, new_fh_init, fromfile, recarray2dict):
        """Setup the hrit file handler for testing."""
        recarray2dict.side_effect = lambda x: x[0]
        new_fh_init.return_value.filename = "filename"
        HRITGOESPrologueFileHandler.filename = "filename"
        HRITGOESPrologueFileHandler.mda = {"total_header_length": 1}
        ret = {}
        the_time = np.array([(32, 24, 18, 146, 19, 50, 121, 153)], dtype=sgs_time)[0]
        for key in ["TCurr", "TCHED", "TCTRL", "TLHED", "TLTRL", "TIPFS",
                    "TINFS", "TISPC", "TIECL", "TIBBC", "TISTR", "TLRAN",
                    "TIIRT", "TIVIT", "TCLMT", "TIONA"]:
            ret[key] = the_time
        ret["SubSatLatitude"] = np.frombuffer(b"\x00\x00\x00\x00", dtype=">i4")[0]
        ret["ReferenceLatitude"] = np.frombuffer(b"\x00\x00\x00\x00", dtype=">i4")[0]
        ret["SubSatLongitude"] = np.frombuffer(b"\x42\x64\x2a\x00", dtype=">i4")[0]
        ret["ReferenceLongitude"] = np.frombuffer(b"\x42\x64\x2a\x00", dtype=">i4")[0]
        ret["ReferenceDistance"] = np.frombuffer(b"\x42\x64\x2a\x00", dtype=">i4")[0]
        ret["SatelliteID"] = 15
        fromfile.return_value = [ret]
        m = mock.mock_open()
        with mock.patch("satpy.readers.goes_imager_hrit.open", m, create=True) as newopen:
            newopen.return_value.__enter__.return_value.seek.return_value = 1
            self.reader = HRITGOESPrologueFileHandler(
                "filename", {"platform_shortname": "GOES15",
                             "start_time": datetime.datetime(2016, 3, 3, 0, 0),
                             "service": "test_service"},
                {"filetype": "info"})

        assert test_pro == self.reader.prologue


class TestHRITGOESFileHandler(unittest.TestCase):
    """Test the HRITFileHandler."""

    @mock.patch("satpy.readers.goes_imager_hrit.HRITFileHandler.__init__")
    def setUp(self, new_fh_init):
        """Set up the hrit file handler for testing."""
        blob = "$HALFTONE:=10\r\n_NAME:=albedo\r\n_UNIT:=percent\r\n0:=0.0\r\n1023:=100.0\r\n".encode()
        mda = {"projection_parameters": {"SSP_longitude": -123.0},
               "spectral_channel_id": 1,
               "image_data_function": blob}
        HRITGOESFileHandler.filename = "filename"
        HRITGOESFileHandler.mda = mda
        self.prologue = mock.MagicMock()
        self.prologue.prologue = test_pro
        self.reader = HRITGOESFileHandler("filename", {}, {}, self.prologue)

    def test_init(self):
        """Test the init."""
        blob = "$HALFTONE:=10\r\n_NAME:=albedo\r\n_UNIT:=percent\r\n0:=0.0\r\n1023:=100.0\r\n".encode()
        mda = {"spectral_channel_id": 1,
               "projection_parameters": {"SSP_longitude": 100.1640625},
               "image_data_function": blob}
        assert self.reader.mda == mda

    @mock.patch("satpy.readers.goes_imager_hrit.HRITFileHandler.get_dataset")
    def test_get_dataset(self, base_get_dataset):
        """Test get_dataset."""
        key = make_dataid(name="CH1", calibration="reflectance")
        base_get_dataset.return_value = DataArray(np.arange(25).reshape(5, 5))
        res = self.reader.get_dataset(key, {})
        expected = np.array([[np.nan, 0.097752, 0.195503, 0.293255, 0.391007],
                             [0.488759, 0.58651, 0.684262, 0.782014, 0.879765],
                             [0.977517, 1.075269, 1.173021, 1.270772, 1.368524],
                             [1.466276, 1.564027, 1.661779, 1.759531, 1.857283],
                             [1.955034, 2.052786, 2.150538, 2.248289, 2.346041]])

        assert np.allclose(res.values, expected, equal_nan=True)
        assert res.attrs["units"] == "%"
        ssp_longitude = self.reader.mda["projection_parameters"]["SSP_longitude"]
        assert res.attrs["orbital_parameters"] == {"projection_longitude": ssp_longitude,
                                                   "projection_latitude": 0.0,
                                                   "projection_altitude": ALTITUDE}

    def test_get_area_def(self):
        """Test getting the area definition."""
        from pyproj import CRS

        self.reader.mda.update({
            "cfac": 10216334,
            "lfac": 10216334,
            "coff": 1408.0,
            "loff": 944.0,
            "number_of_lines": 464,
            "number_of_columns": 2816
        })
        dsid = make_dataid(name="CH1", calibration="reflectance",
                           resolution=3000)
        area = self.reader.get_area_def(dsid)

        expected_crs = CRS(dict(h=ALTITUDE, lon_0=100.1640625, proj="geos", units="m",
                                rf=295.488065897001, a=6378169))
        assert area.crs == expected_crs

        assert area.width == 2816
        assert area.height == 464
        assert area.area_id == "goes-15_goes_imager_fd_3km"
        area_extent_exp = (-5639254.900260435, 1925159.4881528523,
                           5643261.475678028, 3784210.48191544)
        np.testing.assert_allclose(area.area_extent, area_extent_exp)



"""Tests for aws l1b filehandlers."""

import datetime as dt
from enum import Enum

import numpy as np
import pytest

from satpy.tests.reader_tests.conftest import make_fake_angles, make_fake_mwr_lonlats

PLATFORM_NAME = "AWS1"


geo_dims = ["n_scans", "n_fovs", "n_geo_groups"]
geo_size = 10*145*4
shape = (10, 145, 4)
fake_lon_data, fake_lat_data = make_fake_mwr_lonlats(geo_size, geo_dims, shape)
fake_sun_azi_data = make_fake_angles(geo_size, geo_dims, shape)
fake_sun_zen_data = make_fake_angles(geo_size, geo_dims, shape)
fake_sat_azi_data = make_fake_angles(geo_size, geo_dims, shape)
fake_sat_zen_data = make_fake_angles(geo_size, geo_dims, shape)



def test_start_end_time(aws_mwr_handler):
    """Test that start and end times are read correctly."""
    assert aws_mwr_handler.start_time == dt.datetime(2024, 9, 1, 12, 0)
    assert aws_mwr_handler.end_time == dt.datetime(2024, 9, 1, 12, 15)


def test_orbit_number_start_end(aws_mwr_handler):
    """Test that start and end orbit number is read correctly."""
    assert aws_mwr_handler.orbit_start == 9991
    assert aws_mwr_handler.orbit_end == 9992


def test_metadata(aws_mwr_handler):
    """Test that the metadata is read correctly."""
    assert aws_mwr_handler.sensor == "mwr"
    assert aws_mwr_handler.platform_name == PLATFORM_NAME


def test_get_channel_data(aws_mwr_handler, fake_mwr_data_array):
    """Test retrieving the channel data."""
    did = dict(name="1")
    dataset_info = dict(file_key="data/calibration/aws_toa_brightness_temperature")
    expected = fake_mwr_data_array.isel(n_channels=0)
    # mask no_data value
    expected = expected.where(expected != -2147483648)
    # mask outside the valid range
    expected = expected.where(expected <= 700000)
    expected = expected.where(expected >= 0)
    # "calibrate"
    expected = expected * 0.001
    res = aws_mwr_handler.get_dataset(did, dataset_info)
    np.testing.assert_allclose(res, expected)
    assert "x" in res.dims
    assert "y" in res.dims
    assert "orbital_parameters" in res.attrs
    assert res.attrs["orbital_parameters"]["sub_satellite_longitude_end"] == 296.79
    assert res.dims == ("y", "x")
    assert "n_channels" not in res.coords
    assert res.attrs["sensor"] == "mwr"
    assert res.attrs["platform_name"] == PLATFORM_NAME


@pytest.mark.parametrize(("id_name", "file_key", "fake_array"),
                         [("longitude", "data/navigation/aws_lon", fake_lon_data * 1e-4),
                          ("latitude", "data/navigation/aws_lat", fake_lat_data),
                          ])
def test_get_navigation_data(aws_mwr_handler, id_name, file_key, fake_array):
    """Test retrieving the geolocation (lon-lat) data."""
    Horn = Enum("Horn", ["1", "2", "3", "4"])
    did = dict(name=id_name, horn=Horn["1"])
    dataset_info = dict(file_key=file_key, standard_name=id_name)
    res = aws_mwr_handler.get_dataset(did, dataset_info)
    if id_name == "longitude":
        fake_array = fake_array.where(fake_array <= 180, fake_array - 360)

    np.testing.assert_allclose(res, fake_array.isel(n_geo_groups=0))
    assert "x" in res.dims
    assert "y" in res.dims
    assert "orbital_parameters" in res.attrs
    assert res.dims == ("y", "x")
    assert "standard_name" in res.attrs
    assert "n_geo_groups" not in res.coords
    if id_name == "longitude":
        assert res.max() <= 180


@pytest.mark.parametrize(("id_name", "file_key", "fake_array"),
                         [("solar_azimuth_horn1", "data/navigation/aws_solar_azimuth_angle", fake_sun_azi_data),
                          ("solar_zenith_horn1", "data/navigation/aws_solar_zenith_angle", fake_sun_zen_data),
                          ("satellite_azimuth_horn1", "data/navigation/aws_satellite_azimuth_angle", fake_sat_azi_data),
                          ("satellite_zenith_horn1", "data/navigation/aws_satellite_zenith_angle", fake_sat_zen_data)])
def test_get_viewing_geometry_data(aws_mwr_handler, id_name, file_key, fake_array):
    """Test retrieving the angles_data."""
    Horn = Enum("Horn", ["1", "2", "3", "4"])
    dset_id = dict(name=id_name, horn=Horn["1"])

    dataset_info = dict(file_key=file_key, standard_name=id_name)
    res = aws_mwr_handler.get_dataset(dset_id, dataset_info)

    np.testing.assert_allclose(res, fake_array.isel(n_geo_groups=0))
    assert "x" in res.dims
    assert "y" in res.dims
    assert "orbital_parameters" in res.attrs
    assert res.dims == ("y", "x")
    assert "standard_name" in res.attrs
    assert "n_geo_groups" not in res.coords


def test_try_get_data_not_in_file(aws_mwr_handler):
    """Test retrieving a data field that is not available in the file."""
    did = dict(name="toa_brightness_temperature")
    dataset_info = dict(file_key="data/calibration/toa_brightness_temperature")

    match_str = "Dataset toa_brightness_temperature not available or not supported yet!"
    with pytest.raises(NotImplementedError, match=match_str):
        _ = aws_mwr_handler.get_dataset(did, dataset_info)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Test objects and functions in the satpy.config module."""
from __future__ import annotations

import contextlib
import os
import sys
import unittest
from importlib.metadata import EntryPoint
from pathlib import Path
from typing import Callable, Iterator
from unittest import mock

import pytest

import satpy
from satpy import DatasetDict, available_writers
from satpy._config import cached_entry_point
from satpy.composites.config_loader import load_compositor_configs_for_sensors

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmp_path


class TestBuiltinAreas(unittest.TestCase):
    """Test that the builtin areas are all valid."""

    def test_areas_pyproj(self):
        """Test all areas have valid projections with pyproj."""
        import numpy as np
        import pyproj
        import xarray as xr
        from pyresample import parse_area_file
        from pyresample.geometry import SwathDefinition

        from satpy.area import get_area_file

        lons = np.array([[0, 0.1, 0.2], [0.05, 0.15, 0.25]])
        lats = np.array([[0, 0.1, 0.2], [0.05, 0.15, 0.25]])
        lons = xr.DataArray(lons)
        lats = xr.DataArray(lats)
        swath_def = SwathDefinition(lons, lats)
        all_areas = parse_area_file(get_area_file())
        for area_obj in all_areas:
            if hasattr(area_obj, "freeze"):
                try:
                    area_obj = area_obj.freeze(lonslats=swath_def)
                except RuntimeError:
                    # we didn't provide enough info to freeze, hard to guess
                    # in a generic test so just skip this area
                    continue
            _ = pyproj.Proj(area_obj.crs)

    def test_areas_rasterio(self):
        """Test all areas have valid projections with rasterio."""
        try:
            from rasterio.crs import CRS
        except ImportError:
            return unittest.skip("Missing rasterio dependency")
        if not hasattr(CRS, "from_dict"):
            return unittest.skip("RasterIO 1.0+ required")

        import numpy as np
        import xarray as xr
        from pyresample import parse_area_file
        from pyresample.geometry import SwathDefinition

        from satpy.area import get_area_file

        lons = np.array([[0, 0.1, 0.2], [0.05, 0.15, 0.25]])
        lats = np.array([[0, 0.1, 0.2], [0.05, 0.15, 0.25]])
        lons = xr.DataArray(lons)
        lats = xr.DataArray(lats)
        swath_def = SwathDefinition(lons, lats)
        all_areas = parse_area_file(get_area_file())
        for area_obj in all_areas:
            if hasattr(area_obj, "freeze"):
                try:
                    area_obj = area_obj.freeze(lonslats=swath_def)
                except RuntimeError:
                    # we didn't provide enough info to freeze, hard to guess
                    # in a generic test so just skip this area
                    continue
            _ = CRS.from_user_input(area_obj.crs)


@contextlib.contextmanager
def fake_plugin_etc_path(
        tmp_path: Path,
        entry_point_names: dict[str, list[str]],
) -> Iterator[Path]:
    """Create a fake satpy plugin entry point.

    This mocks the necessary methods to trick Satpy into thinking a plugin
    package is installed and has made a satpy plugin available.

    """
    etc_path, entry_points, module_paths = _get_entry_points_and_etc_paths(tmp_path, entry_point_names)
    fake_iter_entry_points = _create_fake_iter_entry_points(entry_points)
    fake_importlib_files = _create_fake_importlib_files(module_paths)
    with mock.patch("satpy._config.entry_points", fake_iter_entry_points), \
            mock.patch("satpy._config.impr_files", fake_importlib_files):
        yield etc_path


def _get_entry_points_and_etc_paths(
        tmp_path: Path,
        entry_point_names: dict[str, list[str]]
) -> tuple[Path, dict[str, list[EntryPoint]], dict[str, Path]]:
    module_path = tmp_path / "satpy_plugin"
    etc_path = module_path / "etc"
    etc_path.mkdir(parents=True, exist_ok=True)
    entry_points: dict[str, list[EntryPoint]] = {}
    entry_point_module_paths: dict[str, Path] = {}
    for ep_group, entry_point_values in entry_point_names.items():
        entry_points[ep_group] = []
        for entry_point_value in entry_point_values:
            parts = [part.strip() for part in entry_point_value.split("=")]
            ep_name = parts[0]
            ep_value = parts[1]
            ep_module = ep_value.split(":")[0].strip()
            ep = EntryPoint(name=ep_name, group=ep_group, value=ep_value)
            entry_points[ep_group].append(ep)
            entry_point_module_paths[ep_module] = module_path
    return etc_path, entry_points, entry_point_module_paths


def _create_fake_iter_entry_points(entry_points: dict[str, list[EntryPoint]]) -> Callable[[], dict[str, EntryPoint]]:
    def _fake_iter_entry_points() -> dict:
        return entry_points
    return _fake_iter_entry_points


def _create_fake_importlib_files(module_paths: dict[str, Path]) -> Callable[[str], Path]:
    def _fake_importlib_files(module_name: str) -> Path:
        return module_paths[module_name]
    return _fake_importlib_files


@pytest.fixture
def fake_composite_plugin_etc_path(tmp_path: Path) -> Iterator[Path]:
    """Create a fake plugin entry point with a fake compositor YAML configuration file."""
    yield from _create_yamlbased_plugin(
        tmp_path,
        "composites",
        "fake_sensor.yaml",
        _write_fake_composite_yaml,
    )


def _write_fake_composite_yaml(yaml_filename: str) -> None:
    with open(yaml_filename, "w") as comps_file:
        comps_file.write("""
    sensor_name: visir/fake_sensor

    composites:
        fake_composite:
            compositor: !!python/name:satpy.composites.core.GenericCompositor
            prerequisites:
            - 3.9
            - 10.8
            - 12.0
            standard_name: fake composite

    """)


@pytest.fixture
def fake_reader_plugin_etc_path(tmp_path: Path) -> Iterator[Path]:
    """Create a fake plugin entry point with a fake reader YAML configuration file."""
    yield from _create_yamlbased_plugin(
        tmp_path,
        "readers",
        "fake_reader.yaml",
        _write_fake_reader_yaml,
    )


def _write_fake_reader_yaml(yaml_filename: str) -> None:
    reader_name = os.path.splitext(os.path.basename(yaml_filename))[0]
    with open(yaml_filename, "w") as comps_file:
        comps_file.write(f"""
reader:
    name: {reader_name}
    sensors: [fake_sensor]
    reader: !!python/name:satpy.readers.core.yaml_reader.FileYAMLReader
datasets: {{}}
""")


@pytest.fixture
def fake_writer_plugin_etc_path(tmp_path: Path) -> Iterator[Path]:
    """Create a fake plugin entry point with a fake writer YAML configuration file."""
    yield from _create_yamlbased_plugin(
        tmp_path,
        "writers",
        "fake_writer.yaml",
        _write_fake_writer_yaml,
    )


def _write_fake_writer_yaml(yaml_filename: str) -> None:
    writer_name = os.path.splitext(os.path.basename(yaml_filename))[0]
    with open(yaml_filename, "w") as comps_file:
        comps_file.write(f"""
writer:
    name: {writer_name}
    writer: !!python/name:satpy.writers.core.base.Writer
""")


@pytest.fixture
def fake_enh_plugin_etc_path(tmp_path: Path) -> Iterator[Path]:
    """Create a fake plugin entry point with a fake enhancement YAML configure files.

    This creates a ``fake_sensor.yaml`` and ``generic.yaml`` enhancement configuration.

    """
    yield from _create_yamlbased_plugin(
        tmp_path,
        "enhancements",
        "fake_sensor.yaml",
        _write_fake_enh_yamls,
    )


def _write_fake_enh_yamls(yaml_filename: str) -> None:
    with open(yaml_filename, "w") as comps_file:
        comps_file.write("""
enhancements:
    some_custom_plugin_enh:
        name: fake_name
        operations:
        - name: stretch
          method: !!python/name:satpy.enhancements.contrast.stretch
          kwargs:
            stretch: crude
            min_stretch: -100.0
            max_stretch: 0.0
""")

    generic_filename = os.path.join(os.path.dirname(yaml_filename), "generic.yaml")
    with open(generic_filename, "w") as comps_file:
        comps_file.write("""
enhancements:
    default:
        operations:
        - name: stretch
          method: !!python/name:satpy.enhancements.contrast.stretch
          kwargs:
            stretch: crude
            min_stretch: -1.0
            max_stretch: 1.0
""")


def _create_yamlbased_plugin(
        tmp_path: Path,
        component_type: str,
        yaml_name: str,
        yaml_func: Callable[[str], None]
) -> Iterator[Path]:
    entry_point_dict = {f"satpy.{component_type}": [f"example_{component_type} = satpy_plugin"]}
    with fake_plugin_etc_path(tmp_path, entry_point_dict) as plugin_etc_path:
        comps_dir = os.path.join(plugin_etc_path, component_type)
        os.makedirs(comps_dir, exist_ok=True)
        comps_filename = os.path.join(comps_dir, yaml_name)
        yaml_func(comps_filename)
        yield plugin_etc_path


class TestPluginsConfigs:
    """Test that plugins are working."""

    @classmethod
    def setup_class(cls):
        """Set up the class of tests with a clean environment."""
        cached_entry_point.cache_clear()

    def teardown_method(self):
        """Tear down the test.

        Make sure we leave every test the way we started.

        """
        cached_entry_point.cache_clear()

    def test_get_plugin_configs(self, fake_composite_plugin_etc_path):
        """Check that the plugin configs are looked for."""
        from satpy._config import get_entry_points_config_dirs

        with satpy.config.set(config_path=[]):
            dirs = get_entry_points_config_dirs("satpy.composites")
            assert dirs == [str(fake_composite_plugin_etc_path)]

    def test_load_entry_point_composite(self, fake_composite_plugin_etc_path):
        """Test that composites can be loaded from plugin entry points."""
        with satpy.config.set(config_path=[]):
            compositors, _ = load_compositor_configs_for_sensors(["fake_sensor"])
            assert "fake_sensor" in compositors
            comp_dict = DatasetDict(compositors["fake_sensor"])
            assert "fake_composite" in comp_dict
            comp_obj = comp_dict["fake_composite"]
            assert comp_obj.attrs["name"] == "fake_composite"
            assert comp_obj.attrs["prerequisites"] == [3.9, 10.8, 12.0]

    @pytest.mark.parametrize("specified_reader", [None, "fake_reader"])
    def test_plugin_reader_configs(self, fake_reader_plugin_etc_path, specified_reader):
        """Test that readers can be loaded from plugin entry points."""
        from satpy.readers.core.config import configs_for_reader
        reader_yaml_path = fake_reader_plugin_etc_path / "readers" / "fake_reader.yaml"
        self._get_and_check_reader_writer_configs(specified_reader, configs_for_reader, reader_yaml_path)

    def test_plugin_reader_available_readers(self, fake_reader_plugin_etc_path):
        """Test that readers can be loaded from plugin entry points."""
        from satpy.readers.core.config import available_readers
        self._check_available_component(available_readers, "fake_reader")

    @pytest.mark.parametrize("specified_writer", [None, "fake_writer"])
    def test_plugin_writer_configs(self, fake_writer_plugin_etc_path, specified_writer):
        """Test that writers can be loaded from plugin entry points."""
        from satpy.writers.core.config import configs_for_writer
        writer_yaml_path = fake_writer_plugin_etc_path / "writers" / "fake_writer.yaml"
        self._get_and_check_reader_writer_configs(specified_writer, configs_for_writer, writer_yaml_path)

    def test_plugin_writer_available_writers(self, fake_writer_plugin_etc_path):
        """Test that readers can be loaded from plugin entry points."""
        self._check_available_component(available_writers, "fake_writer")

    @staticmethod
    def _get_and_check_reader_writer_configs(specified_component, configs_func, exp_yaml):
        with satpy.config.set(config_path=[]):
            configs = list(configs_func(specified_component))
        assert any(str(exp_yaml) in config_list for config_list in configs)

    @staticmethod
    def _check_available_component(available_func, exp_component):
        with satpy.config.set(config_path=[]):
            available_components = available_func()
        assert exp_component in available_components

    @pytest.mark.parametrize(
        ("sensor_name", "exp_result"),
        [
            ("fake_sensor", 1.0),  # uses the sensor specific entry
            ("fake_sensor2", 0.5),  # uses the generic.yaml default
        ]
    )
    def test_plugin_enhancements_generic_sensor(self, fake_enh_plugin_etc_path, sensor_name, exp_result):
        """Test that enhancements from a plugin are available."""
        import dask.array as da
        import numpy as np
        import xarray as xr
        from trollimage.xrimage import XRImage

        from satpy.enhancements.enhancer import Enhancer

        data_arr = xr.DataArray(
            da.zeros((10, 10), dtype=np.float32),
            dims=("y", "x"),
            attrs={
                "sensor": {sensor_name},
                "name": "fake_name",
            })
        img = XRImage(data_arr)

        enh = Enhancer()
        enh.add_sensor_enhancements(data_arr.attrs["sensor"])
        enh.apply(img, **img.data.attrs)

        res_data = img.data.values
        np.testing.assert_allclose(res_data, exp_result)


class TestConfigObject:
    """Test basic functionality of the central config object."""

    def test_custom_config_file(self):
        """Test adding a custom configuration file using SATPY_CONFIG."""
        import tempfile
        from importlib import reload

        import yaml

        import satpy
        my_config_dict = {
            "cache_dir": "/path/to/cache",
        }
        try:
            with tempfile.NamedTemporaryFile(mode="w+t", suffix=".yaml", delete=False) as tfile:
                yaml.dump(my_config_dict, tfile)
                tfile.close()
                with mock.patch.dict("os.environ", {"SATPY_CONFIG": tfile.name}):
                    reload(satpy._config)
                    reload(satpy)
                    assert satpy.config.get("cache_dir") == "/path/to/cache"
        finally:
            os.remove(tfile.name)

    def test_deprecated_env_vars(self):
        """Test that deprecated variables are mapped to new config."""
        from importlib import reload

        import satpy
        old_vars = {
            "PPP_CONFIG_DIR": "/my/ppp/config/dir",
            "SATPY_ANCPATH": "/my/ancpath",
        }

        with mock.patch.dict("os.environ", old_vars):
            reload(satpy._config)
            reload(satpy)
            assert satpy.config.get("data_dir") == "/my/ancpath"
            assert satpy.config.get("config_path") == ["/my/ppp/config/dir"]

    def test_config_path_multiple(self):
        """Test that multiple config paths are accepted."""
        from importlib import reload

        import satpy
        exp_paths, env_paths = _os_specific_multipaths()
        old_vars = {
            "SATPY_CONFIG_PATH": env_paths,
        }

        with mock.patch.dict("os.environ", old_vars):
            reload(satpy._config)
            reload(satpy)
            assert satpy.config.get("config_path") == exp_paths

    def test_config_path_multiple_load(self):
        """Test that config paths from subprocesses load properly.

        Satpy modifies the config path environment variable when it is imported.
        If Satpy is imported again from a subprocess then it should be able to parse this
        modified variable.
        """
        from importlib import reload

        import satpy
        exp_paths, env_paths = _os_specific_multipaths()
        old_vars = {
            "SATPY_CONFIG_PATH": env_paths,
        }

        with mock.patch.dict("os.environ", old_vars):
            # these reloads will update env variable "SATPY_CONFIG_PATH"
            reload(satpy._config)
            reload(satpy)

            # load the updated env variable and parse it again.
            reload(satpy._config)
            reload(satpy)
            assert satpy.config.get("config_path") == exp_paths

    def test_bad_str_config_path(self):
        """Test that a str config path isn't allowed."""
        from importlib import reload

        import satpy
        old_vars = {
            "SATPY_CONFIG_PATH": "/my/configs1",
        }

        # single path from env var still works
        with mock.patch.dict("os.environ", old_vars):
            reload(satpy._config)
            reload(satpy)
            assert satpy.config.get("config_path") == ["/my/configs1"]

        # strings are not allowed, lists are
        with satpy.config.set(config_path="/single/string/paths/are/bad"):
            with pytest.raises(ValueError,
                               match="Satpy config option 'config_path' must be a list, not '<class 'str'>'"):
                satpy._config.get_config_path_safe()

    def test_tmp_dir_is_writable(self):
        """Check that the default temporary directory is writable."""
        import satpy
        assert _is_writable(satpy.config["tmp_dir"])


def test_is_writable():
    """Test writable directory check."""
    assert _is_writable(os.getcwd())
    assert not _is_writable("/foo/bar")


def _is_writable(directory):
    import tempfile
    try:
        with tempfile.TemporaryFile(dir=directory):
            return True
    except OSError:
        return False


def _os_specific_multipaths():
    exp_paths = ["/my/configs1", "/my/configs2", "/my/configs3"]
    if sys.platform.startswith("win"):
        exp_paths = ["C:" + p for p in exp_paths]
    path_str = os.pathsep.join(exp_paths)
    return exp_paths, path_str



# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""ENVISAT MERIS reader.

Sentinel 3 like format: https://earth.esa.int/eogateway/documents/20142/37627/MERIS-Sentinel-3-Like-L1-andL2-PFS.pdf

Default:
    scn = Scene(filenames=my_files, reader='meris_nc_sen3')

References:
    - :func:`xarray.open_dataset`
"""

import logging
from functools import reduce

import numpy as np

from satpy.readers.olci_nc import NCOLCI2, BitFlags, NCOLCIAngles, NCOLCIBase, NCOLCIMeteo

logger = logging.getLogger(__name__)


class NCMERISCal(NCOLCIBase):
    """Dummy class for calibration."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init the meris reader base."""
        super(NCMERISCal, self).__init__(filename, filename_info, filetype_info)
        self.sensor = "meris"


class NCMERISGeo(NCOLCIBase):
    """Dummy class for navigation."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init the meris reader base."""
        super(NCMERISGeo, self).__init__(filename, filename_info, filetype_info)
        self.sensor = "meris"


class NCMERIS2(NCOLCI2):
    """File handler for MERIS l2."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init the file handler."""
        super(NCMERIS2, self).__init__(filename, filename_info, filetype_info)
        self.sensor = "meris"
        self.reflectance_prefix = "M"
        self.reflectance_suffix = "_rho_w"

    def getbitmask(self, wqsf, items=None):
        """Get the bitmask. Experimental default mask."""
        items = items or ["SEA_ICE", "MEGLINT", "HIGHGLINT",
                          "HAZE_OVER_WATER", "WHITECAPS", "AC_FAIL", "WHITE_SCATT",
                          "LOWRW", "HIGHRW", "OUT_OF_RANGE_AAC", "OUT_OF_SCOPE_AAC",
                          "OUT_OF_RANGE_OC_NN", "OUT_OF_SCOPE_OC_NN",
                          "OUT_OF_RANGE_CHL_OC4ME_INPUT", "OUT_OF_RANGE_CHL_OC4ME"]
        bflags = BitFlags(
            wqsf,
            flag_list=["SEA_ICE", "MEGLINT", "HIGHGLINT", "CASE2_S", "CASE2_ANOM",
                       "HAZE_OVER_WATER", "WHITECAPS", "AC_FAIL", "BPAC_ON", "WHITE_SCATT",
                       "LOWRW", "HIGHRW", "OUT_OF_RANGE_AAC", "OUT_OF_SCOPE_AAC",
                       "OUT_OF_RANGE_OC_NN", "OUT_OF_SCOPE_OC_NN",
                       "OUT_OF_RANGE_CHL_OC4ME_INPUT", "OUT_OF_RANGE_CHL_OC4ME"],
        )
        return reduce(np.logical_or, [bflags[item] for item in items])


class NCMERISAngles(NCOLCIAngles):
    """File handler for the MERIS angles."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init the file handler."""
        super(NCMERISAngles, self).__init__(filename, filename_info, filetype_info)
        self.sensor = "meris"


class NCMERISMeteo(NCOLCIMeteo):
    """File handler for the MERIS meteo data."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init the file handler."""
        super(NCMERISMeteo, self).__init__(filename, filename_info, filetype_info)
        self.sensor = "meris"



# Copyright (c) 2024 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the 'oci_l2_bgc' reader."""

import numpy as np
import pytest
from pyresample.geometry import SwathDefinition

from satpy import Scene, available_readers

from .test_seadas_l2 import _create_seadas_chlor_a_netcdf_file

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmp_path_factory


@pytest.fixture(scope="module")
def oci_l2_bgc_netcdf(tmp_path_factory):
    """Create MODIS SEADAS NetCDF file."""
    filename = "PACE_OCI.20211118T175853.L2.OC_BGC.V2_0.NRT.nc4"
    full_path = str(tmp_path_factory.mktemp("oci_l2_bgc") / filename)
    return _create_seadas_chlor_a_netcdf_file(full_path, "PACE", "OCI")


class TestSEADAS:
    """Test the OCI L2 file reader."""

    def test_available_reader(self):
        """Test that OCI L2 reader is available."""
        assert "oci_l2_bgc" in available_readers()

    def test_scene_available_datasets(self, oci_l2_bgc_netcdf):
        """Test that datasets are available."""
        scene = Scene(reader="oci_l2_bgc", filenames=oci_l2_bgc_netcdf)
        available_datasets = scene.all_dataset_names()
        assert len(available_datasets) > 0
        assert "chlor_a" in available_datasets

    @pytest.mark.parametrize("apply_quality_flags", [False, True])
    def test_load_chlor_a(self, oci_l2_bgc_netcdf, apply_quality_flags):
        """Test that we can load 'chlor_a'."""
        reader_kwargs = {"apply_quality_flags": apply_quality_flags}
        scene = Scene(reader="oci_l2_bgc", filenames=oci_l2_bgc_netcdf, reader_kwargs=reader_kwargs)
        scene.load(["chlor_a"])
        data_arr = scene["chlor_a"]
        assert data_arr.dims == ("y", "x")
        assert data_arr.attrs["platform_name"] == "PACE"
        assert data_arr.attrs["sensor"] == {"oci"}
        assert data_arr.attrs["units"] == "mg m^-3"
        assert data_arr.dtype.type == np.float32
        assert isinstance(data_arr.attrs["area"], SwathDefinition)
        assert data_arr.attrs["rows_per_scan"] == 0
        data = data_arr.data.compute()
        if apply_quality_flags:
            assert np.isnan(data[2, 2])
            assert np.count_nonzero(np.isnan(data)) == 1
        else:
            assert np.count_nonzero(np.isnan(data)) == 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests CF-compliant Dataset(s) creation."""
import datetime

import dask.array as da
import numpy as np
import pytest
import xarray as xr
from pyresample import AreaDefinition, create_area_def


class TestCollectCfDataset:
    """Test case for collect_cf_dataset."""

    def test_collect_cf_dataset(self):
        """Test collecting CF datasets from a DataArray objects."""
        from satpy.cf.datasets import _collect_cf_dataset

        geos = AreaDefinition(
            area_id="geos",
            description="geos",
            proj_id="geos",
            projection={"proj": "geos", "h": 35785831., "a": 6378169., "b": 6356583.8},
            width=2, height=2,
            area_extent=[-1, -1, 1, 1])

        # Define test datasets
        data = da.from_array([[1.0, 2.0], [3.0, 4.0]])
        y = [1, 2]
        x = [1, 2]
        time = [1, 2]
        tstart = datetime.datetime(2019, 4, 1, 12, 0)
        tend = datetime.datetime(2019, 4, 1, 12, 15)
        list_dataarrays = [xr.DataArray(data=data, dims=("y", "x"), coords={"y": y, "x": x, "acq_time": ("y", time)},
                                        attrs={"name": "var1", "start_time": tstart, "end_time": tend, "area": geos}),
                           xr.DataArray(data=data, dims=("y", "x"), coords={"y": y, "x": x, "acq_time": ("y", time)},
                                        attrs={"name": "var2", "long_name": "variable 2"})]

        # Collect datasets
        ds = _collect_cf_dataset(list_dataarrays,
                                 include_lonlats=True,
                                 )

        # Test results
        assert len(ds.keys()) == 3
        assert set(ds.keys()) == {"var1", "var2", "geos"}

        da_var1 = ds["var1"]
        da_var2 = ds["var2"]
        assert da_var1.name == "var1"
        assert da_var1.attrs["grid_mapping"] == "geos"
        assert da_var1.attrs["long_name"] == "var1"
        # variable 2
        assert "grid_mapping" not in da_var2.attrs
        assert da_var2.attrs["long_name"] == "variable 2"

    def test_collect_cf_dataset_with_latitude_named_lat(self):
        """Test collecting CF datasets with latitude named lat."""
        from satpy.cf.datasets import _collect_cf_dataset

        data = da.from_array([[75.0, 2.0], [3.0, 4.0]])
        y = [1, 2]
        x = [1, 2]
        geos = AreaDefinition(
            area_id="geos",
            description="geos",
            proj_id="geos",
            projection={"proj": "geos", "h": 35785831., "a": 6378169., "b": 6356583.8},
            width=2, height=2,
            area_extent=[-1, -1, 1, 1])
        datasets = {
            "var1": xr.DataArray(data=data,
                                 dims=("y", "x"),
                                 coords={"y": y, "x": x}),
            "var2": xr.DataArray(data=data,
                                 dims=("y", "x"),
                                 coords={"y": y, "x": x}),
            "lat": xr.DataArray(data=data,
                                dims=("y", "x"),
                                coords={"y": y, "x": x}),
            "lon": xr.DataArray(data=data,
                                dims=("y", "x"),
                                coords={"y": y, "x": x})}
        datasets["lat"].attrs["standard_name"] = "latitude"
        datasets["var1"].attrs["standard_name"] = "dummy"
        datasets["var2"].attrs["standard_name"] = "dummy"
        datasets["var2"].attrs["area"] = geos
        datasets["var1"].attrs["area"] = geos
        datasets["lat"].attrs["name"] = "lat"
        datasets["var1"].attrs["name"] = "var1"
        datasets["var2"].attrs["name"] = "var2"
        datasets["lon"].attrs["name"] = "lon"

        datasets_list = [datasets[key] for key in datasets.keys()]
        datasets_list_no_latlon = [datasets[key] for key in ["var1", "var2"]]

        # Collect datasets
        ds = _collect_cf_dataset(datasets_list, include_lonlats=True)
        ds2 = _collect_cf_dataset(datasets_list_no_latlon, include_lonlats=True)

        # Test results
        assert len(ds.keys()) == 5
        assert set(ds.keys()) == {"var1", "var2", "lon", "lat", "geos"}
        with pytest.raises(KeyError):
            ds["var1"].attrs["latitude"]
        with pytest.raises(KeyError):
            ds["var1"].attrs["longitude"]
        assert ds2["var1"]["latitude"].attrs["name"] == "latitude"
        assert ds2["var1"]["longitude"].attrs["name"] == "longitude"

    def test_geographic_area_coords_attrs(self):
        """Test correct storage for area with lon/lat units."""
        from satpy.cf.datasets import _collect_cf_dataset
        from satpy.tests.utils import make_fake_scene

        scn = make_fake_scene(
            {"ketolysis": np.arange(25.0).reshape(5, 5)},
            daskify=True,
            area=create_area_def("mavas", 4326, shape=(5, 5),
                                 center=(0, 0), resolution=(1, 1)))

        ds = _collect_cf_dataset([scn["ketolysis"]], include_lonlats=False)
        assert ds["ketolysis"].attrs["grid_mapping"] == "mavas"
        assert ds["mavas"].attrs["grid_mapping_name"] == "latitude_longitude"
        assert ds["x"].attrs["units"] == "degrees_east"
        assert ds["y"].attrs["units"] == "degrees_north"
        assert ds["mavas"].attrs["longitude_of_prime_meridian"] == 0.0
        np.testing.assert_allclose(ds["mavas"].attrs["semi_major_axis"], 6378137.0)
        np.testing.assert_allclose(ds["mavas"].attrs["inverse_flattening"], 298.257223563)


class TestCollectCfDatasets:
    """Test case for collect_cf_datasets."""

    def test_empty_collect_cf_datasets(self):
        """Test that if no DataArrays, collect_cf_datasets raise error."""
        from satpy.cf.datasets import collect_cf_datasets

        with pytest.raises(RuntimeError):
            collect_cf_datasets(list_dataarrays=[])



# Copyright (c) 2015-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Resampling in Satpy.

Satpy provides multiple resampling algorithms for resampling geolocated
data to uniform projected grids. The easiest way to perform resampling in
Satpy is through the :class:`~satpy.scene.Scene` object's
:meth:`~satpy.scene.Scene.resample` method. Additional utility functions are
also available to assist in resampling data. Below is more information on
resampling with Satpy as well as links to the relevant API documentation for
available keyword arguments.

Resampling algorithms
---------------------

.. csv-table:: Available Resampling Algorithms
    :header-rows: 1
    :align: center

    "Resampler", "Description", "Related"
    "nearest", "Nearest Neighbor", :class:`~satpy.resample.kdtree.KDTreeResampler`
    "ewa", "Elliptical Weighted Averaging", :class:`~pyresample.ewa.dask_ewa.DaskEWAResampler`
    "ewa_legacy", "Elliptical Weighted Averaging (Legacy)", \
    :class:`~pyresample.ewa._legacy_dask_ewa.LegacyDaskEWAResampler`
    "native", "Native", :class:`~satpy.resample.native.NativeResampler`
    "bilinear", "Bilinear", :class:`~satpy.resample.kdtree.BilinearResampler`
    "bucket_avg", "Average Bucket Resampling", :class:`~satpy.resample.bucket.BucketAvg`
    "bucket_sum", "Sum Bucket Resampling", :class:`~satpy.resample.bucket.BucketSum`
    "bucket_count", "Count Bucket Resampling", :class:`~satpy.resample.bucket.BucketCount`
    "bucket_fraction", "Fraction Bucket Resampling", :class:`~satpy.resample.bucket.BucketFraction`
    "gradient_search", "Gradient Search Resampling", :func:`~pyresample.gradient.create_gradient_search_resampler`

The resampling algorithm used can be specified with the ``resampler`` keyword
argument and defaults to ``nearest``:

.. code-block:: python

    >>> scn = Scene(...)
    >>> euro_scn = scn.resample('euro4', resampler='nearest')

.. warning::

    Some resampling algorithms expect certain forms of data. For example, the
    EWA resampling expects polar-orbiting swath data and prefers if the data
    can be broken in to "scan lines". See the API documentation for a specific
    algorithm for more information.

Resampling for comparison and composites
----------------------------------------

While all the resamplers can be used to put datasets of different resolutions
on to a common area, the 'native' resampler is designed to match datasets to
one resolution in the dataset's original projection. This is extremely useful
when generating composites between bands of different resolutions.

.. code-block:: python

    >>> new_scn = scn.resample(resampler='native')

By default this resamples to the
:meth:`highest resolution area <satpy.scene.Scene.finest_area>` (smallest footprint per
pixel) shared between the loaded datasets. You can easily specify the lowest
resolution area:

.. code-block:: python

    >>> new_scn = scn.resample(scn.coarsest_area(), resampler='native')

Providing an area that is neither the minimum or maximum resolution area
may work, but behavior is currently undefined.

Caching for geostationary data
------------------------------

Satpy will do its best to reuse calculations performed to resample datasets,
but it can only do this for the current processing and will lose this
information when the process/script ends. Some resampling algorithms, like
``nearest`` and ``bilinear``, can benefit by caching intermediate data on disk in the directory
specified by `cache_dir` and using it next time. This is most beneficial with
geostationary satellite data where the locations of the source data and the
target pixels don't change over time.

    >>> new_scn = scn.resample('euro4', cache_dir='/path/to/cache_dir')

See the documentation for specific algorithms to see availability and
limitations of caching for that algorithm.

Create custom area definition
-----------------------------

See :class:`pyresample.geometry.AreaDefinition` for information on creating
areas that can be passed to the resample method::

    >>> from pyresample.geometry import AreaDefinition
    >>> my_area = AreaDefinition(...)
    >>> local_scene = scn.resample(my_area)

Resize area definition in pixels
--------------------------------

Sometimes you may want to create a small image with fixed size in pixels.
For example, to create an image of (y, x) pixels :

    >>> small_scn = scn.resample(scn.finest_area().copy(height=y, width=x), resampler="nearest")


.. warning::

    Be aware that resizing with native resampling (``resampler="native"``) only
    works if the new size is an integer factor of the original input size. For example,
    multiplying the size by 2 or dividing the size by 2. Multiplying by 1.5 would
    not be allowed.


Create dynamic area definition
------------------------------

See :class:`pyresample.geometry.DynamicAreaDefinition` for more information.

Examples coming soon...

Store area definitions
----------------------

Area definitions can be saved to a custom YAML file (see
`pyresample's writing to disk <http://pyresample.readthedocs.io/en/stable/geometry_utils.html#writing-to-disk>`_)
and loaded using pyresample's utility methods
(`pyresample's loading from disk <http://pyresample.readthedocs.io/en/stable/geometry_utils.html#loading-from-disk>`_)::

    >>> from pyresample import load_area
    >>> my_area = load_area('my_areas.yaml', 'my_area')

Or using :func:`satpy.area.get_area_def`, which will search through all
``areas.yaml`` files in your ``SATPY_CONFIG_PATH`` (see :ref:`component_configuration`)::

    >>> from satpy.area import get_area_def
    >>> area_eurol = get_area_def("eurol")

For examples of area definitions, see the file
`etc/areas.yaml <https://github.com/pytroll/satpy/blob/main/satpy/etc/areas.yaml>`_ that is
included with Satpy and where all the area definitions shipped with Satpy are
defined. The section below gives an overview of these area definitions.

Area definitions included in Satpy
----------------------------------

.. include:: /area_def_list.rst

"""
from __future__ import annotations

from typing import Any

from satpy.utils import _import_and_warn_new_location

IMPORT_PATHS = {
    "KDTreeResampler": "satpy.resample.kdtree",
    "BilinearResampler": "satpy.resample.kdtree",
    "NativeResampler": "satpy.resample.native",
    "BucketResamplerBase": "satpy.resample.bucket",
    "BucketAvg": "satpy.resample.bucket",
    "BucketSum": "satpy.resample.bucket",
    "BucketCount": "satpy.resample.bucket",
    "BucketFraction": "satpy.resample.bucket",
    "resample": "satpy.resample.base",
    "prepare_resampler": "satpy.resample.base",
    "resample_dataset": "satpy.resample.base",
    "get_area_file": "satpy.area",
    "get_area_def": "satpy.area",
    "add_xy_coords": "satpy.coords",
    "add_crs_xy_coords": "satpy.coords",
}

def __getattr__(name: str) -> Any:
    new_module = IMPORT_PATHS.get(name)

    if new_module is None:
        raise AttributeError(f"module '{__name__}' has no attribute '{name}'")

    return _import_and_warn_new_location(new_module, name)
